import os
import sys

from OSCRUI import OSCRUI


class Launcher():

    version = '2024.08b210'
    __version__ = '0.2'

    # holds the style of the app
    theme = {
        # general style
        'app': {
            'bg': '#1a1a1a',
            'fg': '#eeeeee',
            'oscr': '#c82934',
            'font': ('Overpass', 11, 'normal'),
            'heading': ('Overpass', 14, 'bold'),
            'subhead': ('Overpass', 12, 'medium'),
            'font-fallback': ('Yu Gothic UI', 'Nirmala UI', 'Microsoft YaHei UI', 'sans-serif'),
            'frame_thickness': 8,
            # this styles every item of the given type
            'style': {
                # scroll bar trough (invisible)
                'QScrollBar': {
                    'background': 'none',
                    'border': 'none',
                    'border-radius': 0,
                    'margin': 0
                },
                'QScrollBar:vertical': {
                    'width': 8,
                },
                'QScrollBar:horizontal': {
                    'height': 8,
                },
                # space above and below the scrollbar handle
                'QScrollBar::add-page, QScrollBar::sub-page': {
                    'background': 'none'
                },
                # scroll bar handle
                'QScrollBar::handle': {
                    'background-color': 'rgba(100,100,100,.75)',
                    'border-radius': 4,
                    'border': 'none'
                },
                # scroll bar arrow buttons
                'QScrollBar::add-line, QScrollBar::sub-line': {
                    'height': 0  # hiding the arrow buttons
                },
                # top left corner of table
                'QTableCornerButton::section': {
                    'background-color': '#1a1a1a'
                },
            }
        },
        # shortcuts, @bg -> means bg in this sub-dictionary
        'defaults': {
            'bg': '#1a1a1a',  # background
            'mbg': '#242424',  # medium background
            'lbg': '#404040',  # light background
            'oscr': '#c82934',  # accent
            'loscr': '#20c82934',  # light accent (12.5% opacity)
            'font': ('Overpass', 11, 'normal'),
            'heading': ('Overpass', 14, 'bold'),
            'subhead': ('Overpass', 12, 'medium'),
            'small_text': ('Overpass', 10, 'normal'),
            'fg': '#eeeeee',  # foreground (usually text)
            'mfg': '#bbbbbb',  # medium foreground
            'bc': '#888888',  # border color
            'bw': 1,  # border width
            'br': 2,  # border radius
            'sep': 2,  # seperator -> width of major seperating lines
            'margin': 10,  # default margin between widgets
            'csp': 5,  # child spacing -> content margin
            'isp': 15,  # item spacing
        },
        # dark frame
        'frame': {
            'background-color': '@bg',
            'border-style': 'none',
            'margin': 0,
            'padding': 0
        },
        # medium frame
        'medium_frame': {
            'background-color': '@mbg',
            'margin': 0,
            'padding': 0
        },
        # light frame
        'light_frame': {
            'background': '@lbg',
            'margin': 0,
            'padding': 0
        },
        # default text (non-button, non-entry, non table)
        'label': {
            'color': '@fg',
            'margin': (3, 0, 3, 0),
            'qproperty-indent': '0',  # disables auto-indent
            'border-style': 'none',
            'font': '@font'
        },
        # heading label
        'label_heading': {
            'color': '@fg',
            'qproperty-indent': '0',
            'border-style': 'none',
            'font': '@heading'
        },
        # label for subheading
        'label_subhead': {
            'color': '@fg',
            'qproperty-indent': '0',
            'border-style': 'none',
            'margin-bottom': 3,
            'font': '@subhead'
        },
        # default button
        'button': {
            'background': 'none',
            'color': '@fg',
            'text-decoration': 'none',
            'border': 'none',
            'border-radius': '@br',
            'margin': (3, 3, 3, 3),
            'padding': (2, 5, 0, 5),
            'font': ('Overpass', 12, 'medium'),
            ':hover': {
                'color': '@fg',
                'border-width': '@bw',
                'border-style': 'solid',
                'border-color': '@oscr'
            },
            ':disabled': {
                'color': '@bc'
            },
            # Tooltip
            '~QToolTip': {
                'background-color': '@mbg',
                'border-style': 'solid',
                'border-color': '@lbg',
                'border-width': '@bw',
                'padding': (0, 0, 0, 0),
                'color': '@fg',
                'font': 'Overpass'
            }
        },
        # button for tab switching
        'tab_button': {
            'background': 'none',
            'color': '@fg',
            'text-decoration': 'none',
            'border-style': 'none',
            'border-width': '@bw',
            'border-color': '@oscr',
            'border-radius': '@br',
            'margin': (3, 3, 3, 3),
            'padding': (2, 5, 0, 5),
            'font': ('Overpass', 15, 'medium'),
            ':hover': {
                'color': '@fg',
                'border-style': 'solid',
            },
            ':checked': {
                'border-style': 'solid',
            }
        },
        # big button (main tab switcher)
        'menu_button': {
            'background': 'none',
            'color': '@fg',
            'text-decoration': 'none',  # removes underline
            'border': 'none',
            'margin': (6, 10, 4, 10),
            # 'margin-left': 10,
            # 'margin-top': 6,
            # 'margin-bottom': 4,
            # 'margin-right': 10,
            'padding': 0,
            'font': ('Overpass', 20, 'bold'),
            ':hover': {
                'text-decoration': 'underline',
                'color': '@fg'
            },
            ':disabled': {
                'color': '@bc'
            }
        },
        # inconspicious button
        'small_button': {
            'background': 'none',
            'border': 'none',
            'border-radius': 3,
            'margin': 0,
            'padding': (2, 0, 2, 0),
            ':hover': {
                'background-color': 'rgba(136,136,136,.2)'
            },
            # Tooltip
            '~QToolTip': {
                'background-color': '@mbg',
                'border-style': 'solid',
                'border-color': '@lbg',
                'border-width': '@bw',
                'padding': (0, 0, 0, 0),
                'margin': (0, 0, 0, 0),
                'color': '@fg',
                'font': 'Overpass'
            }
        },
        # button that holds icon
        'icon_button': {
            'background': 'none',
            'border-width': '@bw',
            'border-style': 'solid',
            'border-color': '@bc',
            'border-radius': 3,
            'margin': 1,
            'padding': (2, 0, 2, 0),
            ':hover': {
                'border-color': '@oscr'
            },
            # Tooltip
            '~QToolTip': {
                'background-color': '@mbg',
                'border-style': 'solid',
                'border-color': '@lbg',
                'border-width': '@bw',
                'padding': (0, 0, 0, 0),
                'color': '@fg',
                'font': 'Overpass'
            }
        },
        # line of user-editable text
        'entry': {
            'background-color': '@lbg',
            'color': '@fg',
            'border-width': '@bw',
            'border-style': 'solid',
            'border-color': '@bc',
            'border-radius': '@br',
            'margin-top': '@csp',
            'font': '@small_text',
            # cursor is inside the line
            ':focus': {
                'border-color': '@oscr'
            }
        },
        # horizontal seperator
        'hr': {
            'background-color': '@oscr',
            'border-style': 'none',
            'margin': (0, 10, 0, 10),
            'height': 2
        },
        # scrollable list of items; ::item refers to the rows
        'listbox': {
            'background-color': '@lbg',
            'color': '@fg',
            'border-width': '@bw',
            'border-style': 'solid',
            'border-color': '@bc',
            'border-radius': '@br',
            'margin': 0,
            'font': '@small_text',
            'outline': '0',  # removes dotted line around clicked item
            '::item': {
                'border-width': '@bw',
                'border-style': 'solid',
                'border-color': '@lbg',
            },
            '::item:selected': {
                'background': 'none',
                'border-width': '@bw',
                'border-style': 'solid',
                'border-color': '@oscr',
                'border-radius': '@br',
            },
            # selected but not the last click of the user
            '::item:selected:!active': {
                'color': '@fg'
            },
            '::item:hover': {
                'background-color': '@loscr',
            },
        },
        # horizontal sliding selector
        'slider': {
            'font': ('Roboto Mono', 11, 'Normal'),
            'color': '@fg',
            'margin-bottom': 3,
            '::groove:horizontal': {
                'border-style': 'none',
                'background-color': '@lbg',
                'border-radius': '@bw',
                'height': 5
            },
            '::handle:horizontal': {
                'border-style': 'solid',
                'border-width': '@bw',
                'border-color': '@bc',
                'border-radius': '@br',
                'background-color': '@bc',
                'width': 8,
                'margin-top': -7,
                'margin-bottom': -7
            },
            '::handle:horizontal:hover': {
                'border-color': '@oscr'
            },
            '::handle:horizontal:pressed': {
                'background-color': '#666666'
            },
        },
        # holds sub-pages
        'tabber': {
            'background': 'none',
            'border': 'none',
            'margin': 0,
            'padding': 0,
            '::pane': {
                'border': 'none',
            }
        },
        # default tabber buttons (hidden)
        'tabber_tab': {
            '::tab': {
                'height': 0,
                'width': 0
            }
        },
        # used in settings
        'toggle_button': {
            'background': 'none',
            'color': '@fg',
            'text-decoration': 'none',
            'border-style': 'solid',
            'border-color': '@bc',
            'border-width': '@bw',
            'border-radius': '@br',
            'margin': (3, 3, 3, 3),
            'padding': (2, 5, 0, 5),
            'font': ('Overpass', 13, 'medium'),
            ':hover': {
                'background-color': '@loscr',
            },
            ':checked': {
                'border-color': '@oscr',
            },
            ':disabled': {
                'color': '@bc'
            }
        },
        # table; ::item refers to the cells,
        #        :alternate is the alternate style -> s.c: table_alternate
        'table': {
            'color': '@fg',
            'background-color': '@bg',
            'border-width': '@bw',
            'border-style': 'solid',
            'border-color': '@bc',
            'gridline-color': 'rgba(0,0,0,0)',  # -> s.c: table_gridline
            'outline': '0',  # removes dotted line around clicked item
            'margin': (0, 0, 10, 0),
            'font': ('Roboto Mono', 12, 'Medium'),
            '::item': {
                'padding': (0, 5, 0, 5),
                'border-width': '@bw',
                'border-style': 'solid',
                'border-color': '@bg',
                'border-right-width': '@bw',
                'border-right-style': 'solid',
                'border-right-color': '@bc',
            },
            '::item:alternate': {
                'padding': (0, 5, 0, 5),
                'background-color': '@mbg',
                'border-width': '@bw',
                'border-style': 'solid',
                'border-color': '@mbg',
                'border-right-width': '@bw',
                'border-right-style': 'solid',
                'border-right-color': '@bc',
            },
            '::item:hover': {
                'background-color': '@loscr',
                'padding': (0, 5, 0, 5)
            },
            '::item:focus': {
                'background-color': '@bg',
                'color': '@fg',
            },
            '::item:selected': {
                'background-color': '@bg',
                'color': '@fg',
                'border': '1px solid #c82934',
            },
            # selected but not the last click of the user
            '::item:alternate:focus': {
                'background-color': '@mbg'
            },
            '::item:alternate:selected': {
                'background-color': '@mbg'
            }
        },
        # heading of the table; ::section refers to the individual buttons
        'table_header': {
            'color': '@bg',
            'background-color': '@mbg',
            'border': 'none',
            'border-bottom-width': '@sep',
            'border-bottom-style': 'solid',
            'border-bottom-color': '@bc',
            'outline': '0',  # removes dotted line around clicked item
            'font': ('Overpass', 12, 'Medium'),
            '::section': {
                'background-color': '@mbg',
                'color': '@fg',
                'padding': (0, 0, 0, 0),
                'border': 'none',
                'margin': 0
            },
            '::section:hover': {
                'background-color': '@loscr'
            },
        },
        # index of the table (vertical header); ::section refers to the individual buttons
        'table_index': {
            'color': '@bg',
            'background-color': '@mbg',
            'border': 'none',
            'border-right-width': '@sep',
            'border-right-style': 'solid',
            'border-right-color': '@bc',
            'outline': '0',  # removes dotted line around clicked item
            '::section': {
                'background-color': '@mbg',
                'color': '@fg',
                'padding': (0, 3, 0, 3),
                'border': 'none',
                'margin': 0
            },
            '::section:hover': {
                'background-color': '@loscr'
            },
        },
        # analysis table; ::item refers to the cells;
        #                 ::branch refers to the space on the left of the rows
        'tree_table': {
            'border': '1px solid #888888',
            'background-color': '@bg',
            'alternate-background-color': '@mbg',
            'color': '@fg',
            'margin': (5, 0, 15, 0),
            'outline': '0',  # removes dotted line around clicked item
            'font': ('Overpass', 12, 'Normal'),
            '::item': {
                'font': ('Roboto Mono', 12, 'Normal'),
                'border-right-width': '@bw',
                'border-right-style': 'solid',
                'border-right-color': '@bc',
                'background-color': 'none',
                'padding': (1, 4, 1, 4)
            },
            '::item:selected, ::item:focus': {
                'border-width': '@bw',
                'border-style': 'solid',
                'border-color': '@oscr',
                'color': '@fg'
            },
            '::item:hover, ::item:alternate:hover': {
                'background-color': '@loscr',
            },
            '::branch:hover': {
                'background-color': '@bg',
                'border': 'none'
            },
            '::branch': {
                'background-color': '@bg'
            },
            # down-pointing arrow
            '::branch:open:has-children': {
                'image': 'url(assets/chevron-down.svg)'
            },
            # right-pointing arrow
            '::branch:closed:has-children': {
                'image': 'url(assets/chevron-right.svg)'
            }
        },
        # header of the analysis table; ::section refers to the individual buttons
        'tree_table_header': {
            'background-color': '@bg',
            'border': 'none',
            'border-bottom-width': '@sep',
            'border-bottom-style': 'solid',
            'border-bottom-color': '@bc',
            'font': '@subhead',
            '::section': {
                'background-color': '@mbg',
                'color': '@fg',
                'border': 'none',
            },
            '::section:hover': {
                'border-width': '@bw',
                'border-style': 'solid',
                'border-color': '@oscr',
            },
        },
        # combo box
        'combobox': {
            'border-style': 'solid',
            'border-width': '@bw',
            'border-color': '@bc',
            'border-radius': '@br',
            'background-color': '@bg',
            'padding': (2, 5, 0, 5),
            'color': '@fg',
            'font': '@subhead',
            '::down-arrow': {
                'image': 'url(assets/thick-chevron-down.svg)',
                'width': '@margin',
            },
            '::drop-down': {
                'border-style': 'none',
                'padding': (2, 2, 2, 2)
            },
            '~QAbstractItemView': {
                'background-color': '@mbg',
                'border-style': 'solid',
                'border-color': '@bc',
                'border-width': '@bw',
                'border-radius': '@br',
                'color': '@fg',
                'outline': '0',
                '::item': {
                    'border-width': '@bw',
                    'border-style': 'solid',
                    'border-color': '@mbg',
                },
                '::item:hover': {
                    'border-color': '@oscr',
                },
            }
        },
        # small window
        'dialog_window': {
            'background-color': '@oscr'
        },
        # frame of the plot widgets
        'plot_widget': {
            'border-style': 'solid',
            'border-width': '@bw',
            'border-color': '@bc',
            'margin': (10, 0, 10, 0),
            'padding': 10,
            'font': ('Overpass', 10, 'bold')
        },
        # undoes all styling of plot_widget to stop inheritance
        'plot_widget_nullifier': {
            'border': 'none',
            'margin': 0,
            'padding': 0,
        },
        # smaller tick font for live parser graph
        'live_plot_widget': {
            'font': ('Overpass', 9, 'medium')
        },
        # holds various properties related to graphing
        'plot': {
            'color_cycler': ('#8f54b4', '#B14D54', '#89B177', '#545DB4', '#C8B74E',
                             '#B45492', '#A27534', '#54A9B4', '#E47B1C', '#BCBCBC'),
        },
        'plot_legend': {
            'font': ('Overpass', 11, 'Medium'),
            'border-style': 'none',
            'padding': 0,
            'margin': 0
        },
        'live_parser': {
            'background-color': '@bg',
            'border-style': 'solid',
            'border-color': '@oscr',
            'border-width': '@sep',
            'border-radius': '@br',
            'margin': 0,
            'padding': 0
        },
        'live_table': {
            'color': '@fg',
            'background-color': '@bg',
            'border-width': '@bw',
            'border-style': 'solid',
            'border-color': '@bc',
            'gridline-color': 'rgba(0,0,0,0)',  # -> s.c: table_gridline
            'outline': '0',  # removes dotted line around clicked item
            'margin': (4, 4, 4, 4),
            'font': ('Roboto Mono', 10, 'Medium'),
            '::item': {
                'padding': (0, 2, 0, 2),
                'border-width': '@bw',
                'border-style': 'solid',
                'border-color': '@bg',
                'border-right-width': '@bw',
                'border-right-style': 'solid',
                'border-right-color': '@bc',
            },
            '::item:alternate': {
                'padding': (0, 2, 0, 2),
                'background-color': '@mbg',
                'border-width': '@bw',
                'border-style': 'solid',
                'border-color': '@mbg',
                'border-right-width': '@bw',
                'border-right-style': 'solid',
                'border-right-color': '@bc',
            }
        },
        # heading of the table; ::section refers to the individual buttons
        'live_table_header': {
            'color': '@bg',
            'background-color': '@mbg',
            'border': 'none',
            'border-bottom-width': '@sep',
            'border-bottom-style': 'solid',
            'border-bottom-color': '@bc',
            'outline': '0',  # removes dotted line around clicked item
            'font': ('Overpass', 10, 'Medium'),
            '::section': {
                'background-color': '@mbg',
                'color': '@fg',
                'padding': (0, 0, 0, 0),
                'border': 'none',
                'margin': 0
            },
        },
        # index of the table (vertical header); ::section refers to the individual buttons
        'live_table_index': {
            'color': '@bg',
            'background-color': '@mbg',
            'border': 'none',
            'border-right-width': '@sep',
            'border-right-style': 'solid',
            'border-right-color': '@bc',
            'outline': '0',  # removes dotted line around clicked item
            '::section': {
                'background-color': '@mbg',
                'color': '@fg',
                'padding': (0, 2, 0, 2),
                'border': 'none',
                'margin': 0
            },
        },
        'resize_handle': {
            'border-style': 'none',
            'background-color': 'none',
            'image': 'url(assets/resize.svg)',
        },
        'splitter': {
            'border': 'none',
            'margin': 0,
            'padding': 0,
            '::handle': {
                'background-color': '@oscr'
            },
            '::handle:pressed': {
                'background-color': '@bc'
            },
            '::handle:vertical': {
                'height': '@bw',
                'margin': (0, 13, 0, 13)
            }
        },
        # other style decisions
        's.c': {
            'sidebar_item_width': 0.2,
            'button_icon_size': 24,
            'table_alternate': True,
            'table_gridline': False,
            'overview_graph_stretch': 10,
            'overview_table_stretch': 3
        }
    }

    @staticmethod
    def base_path() -> str:
        """initialize the base path"""
        try:
            base_path = sys._MEIPASS
        except Exception:
            if getattr(sys, 'frozen', False):
                # The application is frozen
                base_path = os.path.dirname(sys.executable)
            else:
                base_path = os.path.abspath(os.path.dirname(__file__))
        return base_path

    @staticmethod
    def app_config() -> dict:
        config = {
            'minimum_window_width': 1280,
            'minimum_window_height': 720,
            'settings_path': r'/.OSCR_settings.ini',
            'templog_folder_path': r'/~temp_log_files',
            'link_website': 'https://oscr.stobuilds.com',
            'link_github': 'https://github.com/STOCD/OSCR-UI',
            'link_downloads': 'https://github.com/STOCD/OSCR-UI/releases',
            'link_stobuilds': 'https://discord.gg/stobuilds',
            'link_stocd': 'https://github.com/STOCD',
            'live_graph_fields': ('DPS', 'Debuff', 'Attacks-in Share', 'HPS'),
            'ui_scale': 1,
            'live_scale': 1,
            'icon_size': 24,
            'default_settings': {
                'language': 'en',
                'log_path': '',
                'sto_log_path': '',
                'geometry': None,
                'live_geometry': None,
                'live_splitter': None,
                'dmg_columns|0': True,
                'dmg_columns|1': True,
                'dmg_columns|2': True,
                'dmg_columns|3': True,
                'dmg_columns|4': True,
                'dmg_columns|5': True,
                'dmg_columns|6': True,
                'dmg_columns|7': True,
                'dmg_columns|8': True,
                'dmg_columns|9': True,
                'dmg_columns|10': True,
                'dmg_columns|11': True,
                'dmg_columns|12': True,
                'dmg_columns|13': True,
                'dmg_columns|14': True,
                'dmg_columns|15': True,
                'dmg_columns|16': True,
                'dmg_columns|17': True,
                'dmg_columns|18': True,
                'dmg_columns|19': True,
                'dmg_columns|20': True,
                'dmg_columns_length': 21,
                'heal_columns|0': True,
                'heal_columns|1': True,
                'heal_columns|2': True,
                'heal_columns|3': True,
                'heal_columns|4': True,
                'heal_columns|5': True,
                'heal_columns|6': True,
                'heal_columns|7': True,
                'heal_columns|8': True,
                'heal_columns|9': True,
                'heal_columns|10': True,
                'heal_columns|11': True,
                'heal_columns|12': True,
                'heal_columns_length': 13,
                'split_log_after': 480000,
                'seconds_between_combats': 100,
                'excluded_event_ids': ['Autodesc.Combatevent.Falling', ''],
                'graph_resolution': 0.2,
                'combats_to_parse': 10,
                'favorite_ladders': list(),
                'overview_sort_column': 1,
                'overview_sort_order': 'Descending',
                'auto_scan': False,
                'live_columns|0': True,
                'live_columns|1': False,
                'live_columns|2': True,
                'live_columns|3': False,
                'live_columns|4': False,
                'live_columns|5': False,
                'live_columns|6': False,
                'live_parser_opacity': 0.85,
                'live_graph_active': False,
                'live_graph_field': 0,
                'first_overview_tab': 0,
                'log_size_warning': True,
                'ui_scale': 1,
                'live_scale': 1,
                'live_enabled': False,
            }
        }
        return config

    @staticmethod
    def launch():
        args = {}
        exit_code = OSCRUI(
                theme=Launcher.theme, args=args,
                path=Launcher.base_path(), config=Launcher.app_config(),
                versions=(Launcher.__version__, Launcher.version)).run()
        sys.exit(exit_code)


if __name__ == '__main__':
    Launcher.launch()



import os
from re import sub as re_sub


def clean_player_id(id: str) -> str:
    """
    cleans player id and returns handle
    """
    return id[id.find(' ') + 1:-1]


def clean_entity_id(id: str) -> str:
    """
    cleans entity id and returns it
    """
    return re_sub(r'C\[([0-9]+) +?([a-zA-Z_0-9]+)\]', r'\2 \1', id).replace('_', ' ')


def get_entity_num(id: str) -> int:
    """
    gets entity number from entity id
    """
    if not id.startswith('C['):
        return -1
    try:
        return int(re_sub(r'C\[([0-9]+) +?([a-zA-Z_0-9]+)\]', r'\1', id))
    except ValueError:
        return int(re_sub(r'C\[([0-9]+) +?([a-zA-Z_0-9]+)\]_WCB', r'\1', id))
    except TypeError:
        return -1


def format_damage_tree_data(data, column: int) -> str:
    """
    Formats a data point according to TREE_HEADER

    Parameters:
    - :param data: unformatted data point
    - :param column: column of the data point in TREE_HEADER

    :return: formatted data point
    """
    if data == '':
        return ''
    if column == 0:
        if isinstance(data, tuple):
            return ''.join(data)
        return data
    elif column in (3, 5, 6, 7):
        return f'{data * 100:,.2f}%'
    elif column in (1, 2, 4, 13, 14, 15, 16, 17, 18):
        return f'{data:,.2f}'
    elif column in (8, 9, 10, 11, 12, 20, 21):
        return f'{data:,.0f}'
    elif column == 19:
        return f'{data}s'


def format_heal_tree_data(data, column: int) -> str:
    """
    Formats a data point according to HEAL_TREE_HEADER

    Parameters:
    - :param data: unformatted data point
    - :param column: column of the data point in HEAL_TREE_HEADER

    :return: formatted data point
    """
    if data == '':
        return ''
    if column == 0:
        if isinstance(data, tuple):
            return ''.join(data)
        return data
    elif column == 8:
        return f'{data * 100:,.2f}%'
    elif column in (1, 2, 3, 4, 5, 6, 7, 17, 18):
        return f'{data:,.2f}'
    elif column in (9, 10, 12, 13):
        return f'{data:,.0f}'
    elif column == 11:
        return f'{data}s'


def compensate_text(text: str) -> str:
    """
    Unescapes various characters not correctly represented in combatlog files

    Parameters:
    - :param text: str -> text to be cleaned

    :return: cleaned text
    """
    text = text.replace('â€“', '–')
    text = text.replace('Ãœ', 'Ü')
    text = text.replace('Ã¼', 'ü')
    text = text.replace('ÃŸ', 'ß')
    text = text.replace('Ã¶', 'ö')
    text = text.replace('Ã¤', 'ä')
    text = text.replace('â€˜', "'")
    return text


def format_path(path: str):
    if len(path) < 2:
        return path
    path = path.replace(chr(92), '/')
    if path[1] == ':' and path[0] >= 'a' and path[0] <= 'z':
        path = path[0].capitalize() + path[1:]
    if path[-1] != '/':
        if os.path.isdir(path):
            path += '/'
    return path


def format_data(el, integer=False) -> str:
    """
    rounds floats and ints to 2 decimals and sets 1000s seperators, ignores string values

    Parameters:
    - :param el: value to be formatted
    - :param integer: rounds numbers to zero decimal places when True (optional)
    """
    if isinstance(el, (int, float)):
        if not integer:
            return f'{el:,.2f}'
        else:
            return f'{el:,.0f}'
    elif isinstance(el, str):
        el = el.replace('â€“', '–')
        el = el.replace('Ãœ', 'Ü')
        el = el.replace('Ã¼', 'ü')
        el = el.replace('ÃŸ', 'ß')
        el = el.replace('Ã¶', 'ö')
        el = el.replace('Ã¤', 'ä')
        el = el.replace('â€˜', "'")
        return el
    else:
        return str(el)


def format_datetime_str(datetime: str) -> str:
    """
    Formats datetime string into datetime to be displayed.

    Parameters:
    - :param datetime: datetime string; for example "2022-02-23T14:00:27.100000Z"

    :return: formated datetime -> "2022-02-23 14:00:27"
    """
    parts = datetime[:-1].split(':')
    seconds = int(float(parts[-1]))
    return f'{parts[0].replace("T", " ")}:{parts[1]}:{seconds:02d}'


def format_damage_number(num: int) -> str:
    """
    Formats a number:
    - num < 10^3: 123.876 -> 123.9
    - 10^3 < num < 10^6: 123456.789 -> 123.5 k
    - 10^6 < num: 123456789 -> 123.5 M

    Parameters:
    - :param num: number to be formatted
    """
    if num > 1_000_000:
        return f'{num / 1_000_000:.1f} M'
    elif num > 1_000:
        return f'{num / 1000:.1f} k'
    else:
        return f'{num:.1f}'



from .app import OSCRUI

__all__ = [
        'app', 'callbacks', 'datafunctions', 'datamodels', 'displayer', 'iofunctions',
        'leagueconnector', 'style', 'subwindows', 'textedit', 'widgetbuilder', 'widgets']



import os

from PySide6.QtWidgets import QFileDialog, QLineEdit

from OSCR import (
    LIVE_TABLE_HEADER, OSCR, repair_logfile as oscr_repair_logfile, split_log_by_combat,
    split_log_by_lines)

from .iofunctions import browse_path
from .textedit import format_path


def browse_log(self, entry: QLineEdit, translate):
    """
    Callback for browse button.

    Parameters:
    - :param entry: QLineEdit -> path entry line widget
    """
    current_path = entry.text()
    if current_path != '':
        current_path = os.path.dirname(current_path)
    path = self.browse_path(current_path, 'Logfile (*.log);;Any File (*.*)')
    if path != '':
        entry.setText(format_path(path))
        if self.settings.value('auto_scan', type=bool):
            self.analyze_log_callback(translate, path=path, parser_num=1)


def save_combat(self, combat_num: int):
    """
    Callback for save button.

    Parameters:
    - :param combat_num: number of combat in self.combats
    """
    combat = self.parser1.active_combat
    if not combat:
        return
    filename = combat.map
    if combat.difficulty is not None and combat.difficulty != '':
        filename += ' ' + combat.difficulty
    filename += f' {combat.start_time.strftime("%Y-%m-%d %H.%M")}.log'
    base_dir = f'{os.path.dirname(self.entry.text())}/{filename}'
    if not base_dir:
        base_dir = self.app_dir
    path = self.browse_path(base_dir, 'Logfile (*.log);;Any File (*.*)', save=True)
    if path:
        self.parser1.export_combat(combat_num, path)


def navigate_log(self, direction: str):
    """
    Load older or newer combats.

    Parameters:
    - :param direction: "up" -> load newer combats; "down" -> load older combats
    """
    logfile_changed = self.parser1.navigate_log(direction)
    selected_row = self.current_combats.currentRow()
    self.current_combats.clear()
    self.current_combats.addItems(self.parser1.analyzed_combats)
    if logfile_changed:
        self.current_combats.setCurrentRow(0)
        self.current_combat_id = None
        self.analyze_log_callback(0, parser_num=1)
    else:
        self.current_combats.setCurrentRow(selected_row)
    self.widgets.navigate_up_button.setEnabled(self.parser1.navigation_up)
    self.widgets.navigate_down_button.setEnabled(self.parser1.navigation_down)


def switch_analysis_tab(self, tab_index: int):
    """
    Callback for tab switch buttons; switches tab and sets active button.

    Parameters:
    - :param tab_index: index of the tab to switch to
    """
    self.widgets.analysis_tabber.setCurrentIndex(tab_index)
    for index, button in enumerate(self.widgets.analysis_menu_buttons):
        if not index == tab_index:
            button.setChecked(False)
        else:
            button.setChecked(True)


def switch_overview_tab(self, tab_index: int):
    """
    Callback for tab switch buttons; switches tab and sets active button.

    Parameters:
    - :param tab_index: index of the tab to switch to
    """
    self.widgets.overview_tabber.setCurrentIndex(tab_index)
    for index, button in enumerate(self.widgets.overview_menu_buttons):
        if not index == tab_index:
            button.setChecked(False)
        else:
            button.setChecked(True)


def switch_map_tab(self, tab_index: int):
    """
    Callback for tab switch buttons; switches tab and sets active button.

    Parameters:
    - :param tab_index: index of the tab to switch to
    """
    self.widgets.map_tabber.setCurrentIndex(tab_index)
    for index, button in enumerate(self.widgets.map_menu_buttons):
        if not index == tab_index:
            button.setChecked(False)
        else:
            button.setChecked(True)


def switch_main_tab(self, tab_index: int):
    """
    Callback for main tab switch buttons. Switches main and sidebar tabs.

    Parameters:
    - :param tab_index: index of the tab to switch to
    """
    SIDEBAR_TAB_CONVERSION = {
        0: 0,
        1: 0,
        2: 1,
        3: 2
    }
    self.widgets.main_tabber.setCurrentIndex(tab_index)
    self.widgets.sidebar_tabber.setCurrentIndex(SIDEBAR_TAB_CONVERSION[tab_index])
    if tab_index == 0:
        self.widgets.overview_table_button.show()
    else:
        self.widgets.overview_table_button.hide()


def favorite_button_callback(self):
    """
    Adds ladder to / removes ladder from favorites list. Updates settings.
    """
    # Add current ladder to favorites
    current_item = self.widgets.ladder_selector.currentItem()
    if current_item and self.widgets.map_tabber.currentIndex() == 0:
        current_ladder = current_item.text()
        favorite_ladders = self.settings.value('favorite_ladders', type=list)
        if current_ladder not in favorite_ladders:
            favorite_ladders.append(current_ladder)
            self.settings.setValue('favorite_ladders', favorite_ladders)
            self.widgets.favorite_ladder_selector.addItem(current_ladder)
            return

    # Remove current ladder from favorites
    current_item = self.widgets.favorite_ladder_selector.currentItem()
    if current_item:
        current_ladder = current_item.text()
        favorite_ladders = self.settings.value('favorite_ladders', type=list)
        if current_ladder in favorite_ladders:
            favorite_ladders.remove(current_ladder)
            self.settings.setValue('favorite_ladders', favorite_ladders)
            row = self.widgets.favorite_ladder_selector.row(current_item)
            self.widgets.favorite_ladder_selector.takeItem(row)


def set_graph_resolution_setting(self, new_value: int):
    """
    Calculates new_value / 10 and stores it to settings.

    Parameters:
    - :param new_value: data points per second
    """
    try:
        setting_value = round(new_value / 10, 1)
        self.settings.setValue('graph_resolution', setting_value)
        return setting_value
    except (ValueError, ZeroDivisionError):
        return


def set_parser_opacity_setting(self, new_value: int):
    """
    Calculates new_value / 10 and stores it to settings.

    Parameters:
    - :param new_value: 20 times the opacity percentage
    """
    setting_value = f'{new_value / 20:.2f}'
    self.settings.setValue('live_parser_opacity', setting_value)
    return setting_value


def set_ui_scale_setting(self, new_value: int):
    """
    Calculates new_value / 50 and stores it to settings.

    Parameters:
    - :param new_value: 50 times the ui scale percentage
    """
    setting_value = f'{new_value / 50:.2f}'
    self.settings.setValue('ui_scale', setting_value)
    return setting_value


def set_live_scale_setting(self, new_value: int):
    """
    Calculates new_value / 50 and stores it to settings.

    Parameters:
    - :param new_value: 50 times the live scale percentage
    """
    setting_value = f'{new_value / 50:.2f}'
    self.settings.setValue('live_scale', setting_value)
    return setting_value


def set_sto_logpath_setting(self, entry: QLineEdit):
    """
    Formats and stores new logpath to settings.

    Parameters:
    - :param entry: the entry that holds the path
    """
    formatted_path = format_path(entry.text())
    self.settings.setValue('sto_log_path', formatted_path)
    entry.setText(formatted_path)


def browse_sto_logpath(self, entry: QLineEdit):
    """
    Opens prompt to select new logpath and stores it to settings.

    Parameters:
    - :param entry: the entry that holds the path
    """
    current_path = entry.text()
    if not current_path:
        current_path = self.app_dir
    new_path = self.browse_path(os.path.dirname(current_path), 'Logfile (*.log);;Any File (*.*)')
    if new_path:
        formatted_path = format_path(new_path)
        self.settings.setValue('sto_log_path', formatted_path)
        entry.setText(formatted_path)


def auto_split_callback(self, path: str):
    """
    Callback for auto split button
    """
    folder_path = QFileDialog.getExistingDirectory(
            self.window, 'Select Folder', os.path.dirname(path))
    if folder_path:
        split_log_by_lines(
                path, folder_path, self.settings.value('split_log_after', type=int),
                self.settings.value('combat_distance', type=int))


def combat_split_callback(self, path: str, first_num: str, last_num: str):
    """
    Callback for combat split button
    """
    target_path = browse_path(self, path, 'Logfile (*.log);;Any File (*.*)', True)
    if target_path:
        split_log_by_combat(
                path, target_path, int(first_num), int(last_num),
                self.settings.value('seconds_between_combats', type=int),
                self.settings.value('excluded_event_ids', type=list))


def copy_live_data_callback(self):
    """
    Copies the data from the live parser table.
    """
    data_model = self.widgets.live_parser_table.model()
    index_data = data_model._index
    cell_data = data_model._data
    visible_columns = list()
    for i in range(len(LIVE_TABLE_HEADER)):
        visible_columns.append(self.settings.value(f'live_columns|{i}', type=bool))
    output = list()
    for player_name, row in zip(index_data, cell_data):
        output.append(f"`{player_name}`: {row[0]:,.2f} ({row[1]:.1f}s)")
    output = '{ OSCR } DPS (Combat time): ' + ' | '.join(output)
    self.app.clipboard().setText(output)


def expand_overview_table(self):
    """
    Shows the overview table
    """
    self.widgets.overview_table_frame.show()


def collapse_overview_table(self):
    """
    Hides the overview table
    """
    self.widgets.overview_table_frame.hide()


def trim_logfile(self):
    """
    Removes all combats but the most recent one from a logfile
    """
    log_path = os.path.abspath(self.entry.text())
    temp_parser = OSCR(log_path, self.parser_settings)
    if os.path.getsize(log_path) > 125 * 1024 * 1024:
        temp_parser.analyze_massive_log_file()
    else:
        temp_parser.analyze_log_file()
    temp_parser.export_combat(0, log_path)


def repair_logfile(self):
    """
    """
    log_path = os.path.abspath(self.entry.text())
    oscr_repair_logfile(log_path, self.config['templog_folder_path'])



from .translation import init_translation

def init_header_trans(language):
    global _
    _ = init_translation(language)

def get_table_headers():
    return (
        _('Combat Time'), _('DPS'), _('Total Damage'), _('Debuff'), _('Attacks-in Share'),
        _('Taken Damage Share'), _('Damage Share'), _('Max One Hit'), _('Crit Chance'), _('Deaths'), _('Total Heals'),
        _('Heal Share'), _('Heal Crit Chance'), _('Total Damage Taken'), _('Total Hull Damage Taken'),
        _('Total Shield Damage Taken'), _('Total Attacks'), _('Hull Attacks'), _('Attacks-in Number'),
        _('Heal Crit Number'), _('Heal Number'), _('Crit Number'), _('Misses')
    )

def get_tree_headers():
    return (
        '', _('DPS'), _('Total Damage'), _('Debuff'), _('Max One Hit'), _('Crit Chance'), _('Accuracy'), _('Flank Rate'),
        _('Kills'), _('Attacks'), _('Misses'), _('Critical Hits'), _('Flank Hits'), _('Shield Damage'), _('Shield DPS'),
        _('Hull Damage'), _('Hull DPS'), _('Base Damage'), _('Base DPS'), _('Combat Time'), _('Hull Attacks'),
        _('Shield Attacks')  # , _('Shield Resistance')
    )

def get_heal_tree_headers():
    return (
        '', _('HPS'), _('Total Heal'), _('Hull Heal'), _('Hull HPS'), _('Shield Heal'), _('Shield HPS'),
        _('Max One Heal'), _('Crit Chance'), _('Heal Ticks'), _('Critical Heals'), _('Combat Time'), _('Hull Heal Ticks'),
        _('Shield Heal Ticks')
    )

def get_live_table_headers():
    return (
        _('DPS'), _('Combat Time'), _('Debuff'), _('Attacks-in'), _('HPS'), _('Kills'), _('Deaths')
    )

def get_ladder_headers():
    return (
        _('Name'), _('Handle'), _('DPS'), _('Total Damage'), _('Deaths'), _('Combat Time'), _('Date'), _('Max One Hit'), _('Debuff'), _('Highest Damage Ability')
    )



from types import FunctionType, BuiltinFunctionType, MethodType
from typing import Callable

from PySide6.QtCore import QSize, Qt
from PySide6.QtWidgets import QAbstractItemView, QComboBox, QFrame
from PySide6.QtWidgets import QHBoxLayout, QHeaderView, QLabel, QLineEdit
from PySide6.QtWidgets import QPushButton, QSizePolicy, QSlider, QTableView
from PySide6.QtWidgets import QTreeView, QVBoxLayout

from .style import get_style, get_style_class, merge_style, theme_font

CALLABLE = (FunctionType, BuiltinFunctionType, MethodType)

SMINMIN = QSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum)
SMAXMAX = QSizePolicy(QSizePolicy.Policy.Maximum, QSizePolicy.Policy.Maximum)
SMAXMIN = QSizePolicy(QSizePolicy.Policy.Maximum, QSizePolicy.Policy.Minimum)
SMINMAX = QSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Maximum)
SMIXMAX = QSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.Maximum)
SMIXMIN = QSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.Minimum)
SMAX = QSizePolicy.Policy.Maximum
SMIN = QSizePolicy.Policy.Minimum
SEXPAND = QSizePolicy.Policy.Expanding
SFIXED = QSizePolicy.Policy.Fixed

ATOP = Qt.AlignmentFlag.AlignTop
ABOTTOM = Qt.AlignmentFlag.AlignBottom
ARIGHT = Qt.AlignmentFlag.AlignRight
ALEFT = Qt.AlignmentFlag.AlignLeft
ACENTER = Qt.AlignmentFlag.AlignCenter
AVCENTER = Qt.AlignmentFlag.AlignVCenter
AHCENTER = Qt.AlignmentFlag.AlignHCenter

RFIXED = QHeaderView.ResizeMode.Fixed

SMPIXEL = QAbstractItemView.ScrollMode.ScrollPerPixel

SCROLLOFF = Qt.ScrollBarPolicy.ScrollBarAlwaysOff
SCROLLON = Qt.ScrollBarPolicy.ScrollBarAlwaysOn


def create_button(self, text, style: str = 'button', parent=None, style_override={}, toggle=None):
    """
    Creates a button according to style with parent.

    Parameters:
    - :param text: text to be shown on the button
    - :param style: name of the style as in self.theme or style dict
    - :param parent: parent of the button (optional)
    - :param style_override: style dict to override default style (optional)
    - :param toggle: True or False when button should be a toggle button, None when it should be a
    normal button; the bool value indicates the default state of the button

    :return: configured QPushButton
    """
    button = QPushButton(text, parent)
    button.setStyleSheet(get_style_class(self, 'QPushButton', style, style_override))
    if 'font' in style_override:
        button.setFont(theme_font(self, style, style_override['font']))
    else:
        button.setFont(theme_font(self, style))
    button.setCursor(Qt.CursorShape.PointingHandCursor)
    button.setSizePolicy(SMAXMAX)
    if isinstance(toggle, bool):
        button.setCheckable(True)
        button.setChecked(toggle)
    return button


def create_icon_button(
        self, icon, tooltip: str = '', style: str = 'icon_button', parent=None,
        style_override={}, icon_size: tuple = ()) -> QPushButton:
    """
    Creates a button showing an icon according to style with parent.

    Parameters:
    - :param icon: icon to be shown on the button
    - :param tooltip: text to show when the mouse pointer is on the button
    - :param style: name of the style as in self.theme or style dict
    - :param parent: parent of the button (optional)
    - :param style_override: style dict to override default style (optional)
    - :param icon_size: set icon size in case it should be different from the default

    :return: configured QPushButton
    """
    button = QPushButton('', parent)
    button.setIcon(icon)
    if tooltip:
        button.setToolTip(tooltip)
    button.setStyleSheet(get_style_class(self, 'QPushButton', style, style_override))
    if len(icon_size) != 2:
        icon_size = [self.config['icon_size']] * 2
    button.setIconSize(QSize(*icon_size))
    button.setCursor(Qt.CursorShape.PointingHandCursor)
    button.setSizePolicy(SMAXMAX)
    return button


def create_frame(self, parent=None, style='frame', style_override={}, size_policy=None) -> QFrame:
    """
    Creates a frame with default styling and parent

    Parameters:
    - :param parent: parent of the frame (optional)
    - :param style: style dict to override default style (optional)
    - :param size_policy: size policy of the frame (optional)

    :return: configured QFrame
    """
    frame = QFrame(parent)
    frame.setStyleSheet(get_style(self, style, style_override))
    frame.setSizePolicy(size_policy if isinstance(size_policy, QSizePolicy) else SMAXMAX)
    return frame


def create_label(self, text, style: str = 'label', parent=None, style_override={}):
    """
    Creates a label according to style with parent.

    Parameters:
    - :param text: text to be shown on the label
    - :param style: name of the style as in self.theme
    - :param parent: parent of the label (optional)
    - :param style_override: style dict to override default style (optional)

    :return: configured QLabel
    """
    label = QLabel(parent)
    label.setText(text)
    label.setStyleSheet(get_style(self, style, style_override))
    label.setSizePolicy(SMAXMAX)
    if 'font' in style_override:
        label.setFont(theme_font(self, style, style_override['font']))
    else:
        label.setFont(theme_font(self, style))
    return label


def create_button_series(
        self, parent, buttons: dict, style, shape: str = 'row', seperator: str = '', ret=False):
    """
    Creates a row / column of buttons.

    Parameters:
    - :param parent: widget that will contain the buttons
    - :param buttons: dictionary containing button details
        - key "default" contains style override for all buttons (optional)
        - all other keys represent one button, key will be the text on the button; value for the
        key contains dict with details for the specific button (all optional)
            - "callback": callable that will be called on button click
            - "style": individual style override dict
            - "toggle": True or False when button should be a toggle button, None when it should be
            a normal button; the bool value indicates the default state of the button
            - "stretch": stretch value for the button
            - "align": alignment flag for button
    - :param style: key for self.theme -> default style
    - :param shape: row / column
    - :param seperator: string seperator displayed between buttons (optional)

    :return: populated QVBoxlayout / QHBoxlayout
    """
    if 'default' in buttons:
        defaults = merge_style(self, self.theme[style], buttons.pop('default'))
    else:
        defaults = self.theme[style]

    if shape == 'column':
        layout = QVBoxLayout()
    else:
        shape = 'row'
        layout = QHBoxLayout()

    layout.setContentsMargins(0, 0, 0, 0)
    layout.setSpacing(0)

    button_list = []

    if seperator != '':
        sep_style = {
                'color': defaults['color'], 'margin': 0, 'padding': 0, 'background': '#00000000'}

    for i, (name, detail) in enumerate(buttons.items()):
        if 'style' in detail:
            button_style = merge_style(self, defaults, detail['style'])
        else:
            button_style = defaults
        toggle_button = detail['toggle'] if 'toggle' in detail else None
        bt = self.create_button(name, style, parent, button_style, toggle_button)
        if 'callback' in detail and isinstance(detail['callback'], CALLABLE):
            if toggle_button:
                bt.clicked[bool].connect(detail['callback'])
            else:
                bt.clicked.connect(detail['callback'])
        stretch = detail['stretch'] if 'stretch' in detail else 0
        if 'align' in detail:
            layout.addWidget(bt, stretch, detail['align'])
        else:
            layout.addWidget(bt, stretch)
        button_list.append(bt)
        if seperator != '' and i < (len(buttons) - 1):
            sep_label = self.create_label(seperator, 'label', parent, sep_style)
            sep_label.setSizePolicy(SMAXMIN)
            layout.addWidget(sep_label)

    if ret:
        return layout, button_list
    else:
        return layout


def create_combo_box(self, parent, style: str = 'combobox', style_override: dict = {}) -> QComboBox:
    """
    Creates a combobox with given style and returns it.

    Parameters:
    - :param parent: parent of the combo box
    - :param style: key for self.theme -> default style
    - :param style_override: style dict to override default style

    :return: styled QCombobox
    """
    combo_box = QComboBox(parent)
    combo_box.setStyleSheet(get_style_class(self, 'QComboBox', style, style_override))
    if 'font' in style_override:
        combo_box.setFont(theme_font(self, style, style_override['font']))
    else:
        combo_box.setFont(theme_font(self, style))
    combo_box.setSizePolicy(SMINMAX)
    combo_box.setCursor(Qt.CursorShape.PointingHandCursor)
    combo_box.view().setCursor(Qt.CursorShape.PointingHandCursor)
    return combo_box


def create_entry(
        self, default_value='', validator=None, style: str = 'entry',
        style_override: dict = {}, placeholder='') -> QLineEdit:
    """
    Creates an entry widget and styles it.

    Parameters:
    - :param default_value: default value for the entry
    - :param validator: validator to validate entered characters against
    - :param style: key for self.theme -> default style
    - :param style_override: style dict to override default style
    - :param placeholder: placeholder shown when entry is empty

    :return: styled QLineEdit
    """
    entry = QLineEdit(default_value)
    entry.setValidator(validator)
    entry.setPlaceholderText(placeholder)
    entry.setStyleSheet(get_style_class(self, 'QLineEdit', style, style_override))
    if 'font' in style_override:
        entry.setFont(theme_font(self, style, style_override['font']))
    else:
        entry.setFont(theme_font(self, style))
    entry.setCursor(Qt.CursorShape.IBeamCursor)
    entry.setSizePolicy(SMAXMAX)
    return entry


def create_annotated_slider(
        self, default_value: int = 1, min: int = 0, max: int = 3,
        style: str = 'slider', style_override_slider: dict = {}, style_override_label: dict = {},
        callback: Callable = lambda v: v) -> QHBoxLayout:
    """
    Creates Slider with label to display the current value.

    Parameters:
    - :param default_value: start value for the slider
    - :param min: lowest value of the slider
    - :param max: highest value of the slider
    - :param style: key for self.theme -> default style
    - :param style_override_slider: style dict to override default style
    - :param style_override_label: style dict to override default style
    - :param callback: callable to be attached to the valueChanged signal of the slider; will be
    passed value the slider was moved to; must return value that the label should be set to

    :return: layout with slider
    """
    def label_updater(new_value):
        if isinstance(callback, CALLABLE):
            new_text = callback(new_value)
            slider_label.setText(str(new_text))

    layout = QHBoxLayout()
    layout.setContentsMargins(0, 0, 0, 3)
    layout.setSpacing(self.theme['defaults']['margin'])
    slider_label = create_label(
            self, '', style, style_override=style_override_label)
    layout.addWidget(slider_label, alignment=AVCENTER)
    slider = QSlider(Qt.Orientation.Horizontal)
    slider.setRange(min, max)
    slider.setSingleStep(1)
    slider.setPageStep(1)
    slider.setValue(default_value)
    slider.setTickPosition(QSlider.TickPosition.NoTicks)
    slider.setFocusPolicy(Qt.FocusPolicy.WheelFocus)
    slider.setSizePolicy(SMINMAX)
    slider.setStyleSheet(get_style_class(self, 'QSlider', style, style_override_slider))
    slider.valueChanged.connect(label_updater)
    layout.addWidget(slider, stretch=1, alignment=AVCENTER)
    label_updater(default_value)
    return layout


def resize_tree_table(tree: QTreeView):
    """
    Resizes the columns of the given tree table to fit its contents

    Parameters:
    - :param tree: QTreeView -> tree to be resized
    """
    for col in range(tree.header().count()):
        width = max(tree.sizeHintForColumn(col), tree.header().sectionSizeHint(col)) + 5
        tree.header().resizeSection(col, width)


def create_analysis_table(self, parent, widget) -> QTreeView:
    """
    Creates and returns a QTreeView with parent, styled according to widget.

    Parameters:
    - :param parent: parent of the table
    - :param widget: style key for the table

    :return: configured QTreeView
    """
    table = QTreeView(parent)
    table.setStyleSheet(get_style_class(self, 'QTreeView', widget))
    table.setSizePolicy(SMINMIN)
    table.setAlternatingRowColors(True)
    table.setHorizontalScrollMode(SMPIXEL)
    table.setVerticalScrollMode(SMPIXEL)
    table.setSortingEnabled(True)
    table.setEditTriggers(QAbstractItemView.EditTrigger.NoEditTriggers)
    table.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)
    table.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectItems)
    table.header().setStyleSheet(get_style_class(self, 'QHeaderView', 'tree_table_header'))
    table.header().setSectionResizeMode(RFIXED)
    table.header().setMinimumSectionSize(1)
    table.header().setSectionsClickable(True)
    table.header().setStretchLastSection(False)
    table.header().setSortIndicatorShown(False)
    table.expanded.connect(lambda: resize_tree_table(table))
    table.collapsed.connect(lambda: resize_tree_table(table))
    return table


def style_table(self, table: QTableView, style_override: dict = {}, single_row_selection=False):
    """
    Styles the given table.

    Parameters:
    - :param table: table to be styled
    - :param style_override: style override for table
    - :param single_row_selection: True when only one row should be selectable at once
    """
    table.setAlternatingRowColors(self.theme['s.c']['table_alternate'])
    table.setShowGrid(self.theme['s.c']['table_gridline'])
    table.setSortingEnabled(True)
    table.setStyleSheet(get_style_class(self, 'QTableView', 'table', style_override))
    table.setHorizontalScrollMode(SMPIXEL)
    table.setVerticalScrollMode(SMPIXEL)
    table.horizontalHeader().setStyleSheet(get_style_class(self, 'QHeaderView', 'table_header'))
    table.verticalHeader().setStyleSheet(get_style_class(self, 'QHeaderView', 'table_index'))
    table.verticalHeader().setMinimumHeight(1)
    table.verticalHeader().setDefaultSectionSize(1)
    table.resizeColumnsToContents()
    table.resizeRowsToContents()
    table.horizontalHeader().setSortIndicatorShown(False)
    table.horizontalHeader().setSectionResizeMode(RFIXED)
    table.verticalHeader().setSectionResizeMode(RFIXED)
    table.setSizePolicy(SMINMIN)
    if single_row_selection:
        table.setSelectionMode(QAbstractItemView.SelectionMode.SingleSelection)
        table.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)



from typing import Callable, Iterable

import numpy as np
from pyqtgraph import BarGraphItem, mkPen, PlotWidget, setConfigOptions
from PySide6.QtWidgets import QFrame, QHBoxLayout, QLabel, QTableView, QVBoxLayout, QWidget
from PySide6.QtCore import Qt, Slot

from .headers import get_table_headers
from OSCR.combat import Combat

from .datamodels import OverviewTableModel, SortingProxy
from .widgetbuilder import ACENTER, AVCENTER, SMINMIN, SMIXMAX
from .widgetbuilder import create_frame, create_label, style_table
from .widgets import CustomPlotAxis
from .style import get_style, theme_font

setConfigOptions(antialias=True)


def setup_plot(plot_function: Callable) -> Callable:
    '''
    sets up Plot item and puts it into layout
    '''
    def plot_wrapper(self, data, time_reference=None):
        plot_widget = PlotWidget()
        plot_widget.setAxisItems({'left': CustomPlotAxis('left')})
        plot_widget.setAxisItems({'bottom': CustomPlotAxis('bottom')})
        plot_widget.setStyleSheet(get_style(self, 'plot_widget_nullifier'))
        plot_widget.setBackground(None)
        plot_widget.setMouseEnabled(False, False)
        plot_widget.setMenuEnabled(False)
        plot_widget.hideButtons()
        plot_widget.setDefaultPadding(padding=0)
        left_axis = plot_widget.getAxis('left')
        left_axis.setTickFont(theme_font(self, 'plot_widget'))
        left_axis.setTextPen(color=self.theme['defaults']['fg'])
        bottom_axis = plot_widget.getAxis('bottom')
        bottom_axis.setTickFont(theme_font(self, 'plot_widget'))
        bottom_axis.setTextPen(color=self.theme['defaults']['fg'])

        if time_reference is None:
            legend_data = plot_function(self, data, plot_widget)
        else:
            legend_data = plot_function(self, data, time_reference, plot_widget)

        inner_layout = QVBoxLayout()
        inner_layout.setContentsMargins(0, 0, 0, 0)
        inner_layout.setSpacing(self.theme['defaults']['isp'])
        inner_layout.addWidget(plot_widget)
        if legend_data is not None:
            legend_frame = create_legend(self, legend_data)
            inner_layout.addWidget(legend_frame, alignment=ACENTER)
        frame = create_frame(self, None, 'plot_widget', size_policy=SMINMIN)
        frame.setLayout(inner_layout)
        outer_layout = QVBoxLayout()
        outer_layout.setContentsMargins(0, 0, 0, 0)
        outer_layout.addWidget(frame)
        return outer_layout
    return plot_wrapper


def extract_overview_data(combat: Combat) -> tuple:
    '''
    converts dictionary containing player data to table data for the front page
    '''
    table = list()

    DPS_graph_data = dict()
    DMG_graph_data = dict()
    graph_time = dict()

    for player in combat.player_dict.values():
        table.append((*player,))

        DPS_graph_data[player.handle] = player.DPS_graph_data
        DMG_graph_data[player.handle] = player.DMG_graph_data
        graph_time[player.handle] = player.graph_time
    table.sort(key=lambda x: x[0])
    return (graph_time, DPS_graph_data, DMG_graph_data, table)


def create_overview(self):
    """
    creates the main Parse Overview including graphs and table
    """
    # clear graph frames
    for frame in self.widgets.overview_tab_frames:
        if frame.layout():
            QWidget().setLayout(frame.layout())
    if self.widgets.overview_table_frame.layout():
        QWidget().setLayout(self.widgets.overview_table_frame.layout())

    time_data, DPS_graph_data, DMG_graph_data, current_table = extract_overview_data(
            self.parser1.active_combat)

    line_layout = create_line_graph(self, DPS_graph_data, time_data)
    self.widgets.overview_tab_frames[1].setLayout(line_layout)

    group_bar_layout = create_grouped_bar_plot(self, DMG_graph_data, time_data)
    self.widgets.overview_tab_frames[2].setLayout(group_bar_layout)

    bar_layout = create_horizontal_bar_graph(self, current_table)
    self.widgets.overview_tab_frames[0].setLayout(bar_layout)

    table_layout = QVBoxLayout()
    table_layout.setContentsMargins(0, 0, 0, 0)
    table = create_overview_table(self, current_table)
    table_layout.addWidget(table)
    self.widgets.overview_table_frame.setLayout(table_layout)
    table.resizeColumnsToContents()


@setup_plot
def create_grouped_bar_plot(
        self, data: dict[str, tuple], time_reference: dict[str, tuple],
        bar_widget: PlotWidget) -> QVBoxLayout:
    """
    Creates a bar plot with grouped bars.

    Parameters:
    - :param data: dictionary containing the data to be plotted
    - :param time_reference: contains the time values for the data points
    - :param bar_widget: bar widget that will be plotted to (supplied by decorator)

    :return: layout containing the graph
    """
    bottom_axis = bar_widget.getAxis('bottom')
    bottom_axis.unit = 's'
    legend_data = list()

    group_width = 0.18
    player_num = len(data)
    if player_num == 0:
        return
    bar_width = group_width / player_num
    relative_bar_positions = np.linspace(0 + bar_width / 2, group_width - bar_width / 2, player_num)
    bar_position_offsets = relative_bar_positions - np.median(relative_bar_positions)

    zipper = zip(data.items(), self.theme['plot']['color_cycler'], bar_position_offsets)
    for (player, graph_data), color, offset in zipper:
        if player in time_reference:
            time_data = np.subtract(time_reference[player], offset)
            bars = BarGraphItem(
                    x=time_data, width=bar_width, height=graph_data, brush=color, pen=None,
                    name=player)
            bar_widget.addItem(bars)
            legend_data.append((color, player))
    return legend_data


@setup_plot
def create_horizontal_bar_graph(self, table: list[list], bar_widget: PlotWidget) -> QVBoxLayout:
    """
    Creates bar plot from table and returns layout in which the graph was inserted.

    Parameters:
    - :param table: overview table as generated by the parser
    - :param bar_widget: bar widget that will be plotted to (supplied by decorator)

    :return: layout containing the graph (returned by decorator)
    """
    left_axis = bar_widget.getAxis('left')
    left_axis.setTickFont(theme_font(self, 'app'))
    bar_widget.setDefaultPadding(padding=0.01)

    table.sort(key=lambda line: line[3], reverse=True)
    y_annotations = (tuple((index + 1, line[0] + line[1]) for index, line in enumerate(table)),)
    bar_widget.getAxis('left').setTicks(y_annotations)
    x = tuple(line[3] for line in table)
    y = tuple(range(1, len(x) + 1))
    bar_widget.setXRange(0, max(x) * 1.05, padding=0)
    bars = BarGraphItem(
            x0=0, y=y, height=0.75, width=x, brush=self.theme['defaults']['mfg'], pen=None)
    bar_widget.addItem(bars)


@setup_plot
def create_line_graph(
        self, data: dict[str, tuple], time_reference: dict[str, tuple],
        graph_widget: PlotWidget) -> QVBoxLayout:
    """
    Creates line plot from data and returns layout that countins the plot.

    Parameters:
    - :param data: dictionary containing the data to be plotted
    - :param time_reference: contains the time values for the data points
    - :param graph_widget: graph widget that will be plotted to (supplied by decorator)

    :return: layout containing the graph (returned by decorator)
    """
    bottom_axis = graph_widget.getAxis('bottom')
    bottom_axis.unit = 's'
    legend_data = list()

    for (player, graph_data), color in zip(data.items(), self.theme['plot']['color_cycler']):
        if player in time_reference:
            time_data = time_reference[player]
            graph_widget.plot(time_data, graph_data, pen=mkPen(color, width=1.5), name=player)
            legend_data.append((color, player))
    return legend_data


def create_legend(self, colors_and_names: Iterable[tuple]) -> QFrame:
    """
    Creates Legend from color / name pairs and returns frame containing it.

    Parameters:
    - :param colors_and_names: Iterable containing color / name pairs : [('#9f9f00', 'Line 1'),
    ('#0000ff', 'Line 2'), (...), ...]

    :return: frame containing the legend
    """
    frame = create_frame(self, style='plot_legend')
    upper_frame = create_frame(self, style='plot_legend')
    lower_frame = create_frame(self, style='plot_legend')
    frame_layout = QVBoxLayout()
    upper_layout = QHBoxLayout()
    lower_layout = QHBoxLayout()
    margin = self.theme['defaults']['margin']
    frame_layout.setContentsMargins(0, 0, 0, 0)
    frame_layout.setSpacing(margin)
    upper_layout.setContentsMargins(0, 0, 0, 0)
    upper_layout.setSpacing(2 * margin)
    lower_layout.setContentsMargins(0, 0, 0, 0)
    lower_layout.setSpacing(2 * margin)
    second_row = False
    for num, (color, name) in enumerate(colors_and_names, 1):
        legend_item = create_legend_item(self, color, name)
        if num <= 5:
            upper_layout.addWidget(legend_item)
        else:
            second_row = True
            lower_layout.addWidget(legend_item)
    upper_frame.setLayout(upper_layout)
    frame_layout.addWidget(upper_frame, alignment=ACENTER)
    if second_row:
        lower_frame.setLayout(lower_layout)
        frame_layout.addWidget(lower_frame, alignment=ACENTER)
    frame.setLayout(frame_layout)
    return frame


def create_legend_item(self, color: str, name: str) -> QFrame:
    """
    Creates a colored patch next to a label inside a frame

    Parameters:
    - :param color: patch color
    - :param name: text of the label

    :return: frame containing the legend item
    """
    frame = create_frame(self, style='plot_legend')
    layout = QHBoxLayout()
    layout.setContentsMargins(0, 0, 0, 0)
    layout.setSpacing(self.theme['defaults']['margin'])
    colored_patch = QLabel()
    colored_patch.setStyleSheet(get_style(self, 'plot_legend', {'background-color': color}))
    patch_height = self.theme['app']['frame_thickness']
    colored_patch.setFixedSize(2 * patch_height, patch_height)
    layout.addWidget(colored_patch, alignment=AVCENTER)
    label = create_label(
            self, name, 'label', style_override={'font': self.theme['plot_legend']['font']})
    layout.addWidget(label)
    frame.setLayout(layout)
    return frame


def create_overview_table(self, table_data) -> QTableView:
    """
    Creates the overview table and returns it.

    :return: Overview Table
    """
    table_cell_data = tuple(tuple(line[2:]) for line in table_data)
    table_index = tuple(line[0] + line[1] for line in table_data)
    model = OverviewTableModel(
            table_cell_data, get_table_headers(), table_index, self.theme_font('table_header'),
            self.theme_font('table'))
    sort = SortingProxy()
    sort.setSourceModel(model)
    table = QTableView(self.widgets.overview_tab_frames[0])
    table.setModel(sort)
    style_table(self, table)
    if self.settings.value('overview_sort_order') == 'Descending':
        sort_order = Qt.SortOrder.AscendingOrder
    else:
        sort_order = Qt.SortOrder.DescendingOrder
    table.sortByColumn(self.settings.value('overview_sort_column', type=int), sort_order)
    return table


def create_live_graph(self) -> tuple[QFrame, list]:
    """
    Creates and styles live graph.

    :return: Frame containing the graph and list of curves that will be used to plot the data
    """
    plot_widget = PlotWidget()
    plot_widget.setAxisItems({'left': CustomPlotAxis('left', compressed=True)})
    plot_widget.setAxisItems({'bottom': CustomPlotAxis('bottom', unit='s', no_labels=True)})
    plot_widget.setStyleSheet(get_style(self, 'plot_widget_nullifier'))
    plot_widget.setBackground(None)
    plot_widget.setMouseEnabled(False, False)
    plot_widget.setMenuEnabled(False)
    plot_widget.hideButtons()
    plot_widget.setDefaultPadding(padding=0)
    plot_widget.setXRange(-14, 0, padding=0)
    left_axis = plot_widget.getAxis('left')
    left_axis.setTickFont(theme_font(self, 'live_plot_widget'))
    left_axis.setTextPen(color=self.theme['defaults']['fg'])
    # left_axis.setTickDensity(0.1)
    bottom_axis = plot_widget.getAxis('bottom')
    bottom_axis.setTickFont(theme_font(self, 'plot_widget'))
    bottom_axis.setTextPen(color=self.theme['defaults']['fg'])

    curves = list()
    for color_index in range(5):
        color = self.theme['plot']['color_cycler'][color_index]
        curves.append(plot_widget.plot([0], [0], pen=mkPen(color, width=1)))

    frame = create_frame(self, None, 'plot_widget', size_policy=SMIXMAX, style_override={
            'margin': 4, 'padding': 2})
    frame.setMinimumWidth(self.sidebar_item_width * 0.25)
    frame.setMinimumHeight(self.sidebar_item_width * 0.25)
    layout = QHBoxLayout()
    layout.setContentsMargins(0, 0, 0, 0)
    layout.addWidget(plot_widget, stretch=1)
    frame.setLayout(layout)
    return frame, curves


def update_live_display(
        self, data: dict, graph_active: bool = False, graph_data_buffer: list = [],
        graph_data_field: int = 0):
    """
    Updates display of live parser to show the new data.

    Parameters:
    - :param data: dictionary containing the new data
    - :param graph_active: Set to True to update the graph as well
    - :param graph_data_buffer: contains the past graph data
    """
    index = list()
    cells = list()
    curves = list()
    for player, player_data in data.items():
        index.append(player)
        cells.append(list(player_data.values()))
    if graph_active:
        if len(graph_data_buffer) == 0:
            graph_data_buffer.extend(([0] * 15, [0] * 15, [0] * 15, [0] * 15, [0] * 15))
        zipper = zip(graph_data_buffer, cells, self.widgets.live_parser_curves)
        for buffer_item, player_data, curve in zipper:
            buffer_item.pop(0)
            buffer_item.append(player_data[graph_data_field])
            curves.append((curve, buffer_item))
        if len(curves) > 0:
            self.live_parser_window.update_graph.emit(curves)

    if len(index) > 0 and len(cells) > 0:
        self.live_parser_window.update_table.emit((index, cells))


@Slot()
def update_live_table(self, data: tuple):
    """
    Updates the table of the live parser with the supplied data

    Parameters:
    - :param data: tuple containing two lists, that contain the index and cell values respectively
    """
    table = self.widgets.live_parser_table
    table.model().replace_data(*data)
    table.resizeColumnsToContents()
    table.resizeRowsToContents()


@Slot()
def update_live_graph(curve_data: list):
    """
    Updates the graph of the live parser with the supplied data

    Parameters:
    - :param curve_data: list containing pairs of curve items and data lists; curve items will be
    updated with the data
    """
    time_data = list(range(-14, 1))
    for curve, data_points in curve_data:
        curve.setData(time_data, data_points)



import json
import os
import re
import shutil
import sys
import webbrowser

from PySide6.QtWidgets import QFileDialog
from PySide6.QtGui import QIcon

# --------------------------------------------------------------------------------------------------
# object methods
# --------------------------------------------------------------------------------------------------


def browse_path(self, default_path: str = None, types: str = 'Any File (*.*)', save=False) -> str:
    """
    Opens file dialog prompting the user to select a file.

    Parameters:
    - :param default_path: path that the file dialog opens at
    - :param types: string containing all file extensions and their respective names that are
    allowed.
    Format: "<name of file type> (*.<extension>);;<name of file type> (*.<extension>);; [...]"
    Example: "Logfile (*.log);;Any File (*.*)"
    """
    if default_path is None or default_path == '':
        default_path = self.app_dir
    default_path = os.path.abspath(default_path)
    if not os.path.exists(os.path.dirname(default_path)):
        default_path = self.app_dir
    if save:
        f = QFileDialog.getSaveFileName(self.window, 'Save Log', default_path, types)[0]
    else:
        f = QFileDialog.getOpenFileName(self.window, 'Open Log', default_path, types)[0]
        if not os.path.exists(f):
            return ''
    return f

# --------------------------------------------------------------------------------------------------
# static functions
# --------------------------------------------------------------------------------------------------


def get_asset_path(asset_name: str, app_directory: str) -> str:
    """
    returns the absolute path to a file in the asset folder

    Parameters:
    - :param asset_name: filename of the asset
    - :param app_directory: absolute path to app directory
    """
    fp = os.path.join(app_directory, 'assets', asset_name)
    if os.path.exists(fp):
        return fp
    else:
        return ''


def load_icon(filename: str, app_directory: str) -> QIcon:
    """
    Loads icon from path and returns it.

    Parameters:
    - :param path: path to icon
    - :param app_directory: absolute path to the app directory
    """
    return QIcon(get_asset_path(filename, app_directory))


def load_icon_series(icons: dict, app_directory: str) -> dict[str, QIcon]:
    """
    Loads multiple icons and returns dictionary containing the icons.

    Parameters:
    - :param icons: contains icons; format: {"<icon_name>": "<icon_path>", [...]}
    - :param app_directory: absolute path to app directory

    :return: dictionary containing icons; format: {"<icon_name>": "<icon>", [...]}
    """
    asset_path = os.path.join(app_directory, 'assets')
    icon_dict = dict()
    for icon_name, file_name in icons.items():
        icon_dict[icon_name] = QIcon(os.path.join(asset_path, file_name))
    return icon_dict


def open_link(link: str = ''):
    """
    Opens provided link
    """
    if link:
        webbrowser.open(link, new=2, autoraise=True)


def fetch_json(path: str) -> dict | list:
    """
    Fetches json from path and returns dictionary.

    Parameters:
    - :param path: path to json file
    """
    if not (os.path.exists(path) and os.path.isfile(path) and os.path.isabs(path)):
        raise FileNotFoundError('Invalid Path')
    with open(path, 'r', encoding='utf-8') as file:
        data = json.load(file)
    return data


def store_json(data: dict | list, path: str):
    """
    Stores data to json file at path. Overwrites file at target location.

    Paramters:
    - :param data: dictionary or list that should be stored
    - :param path: target location; must be absolute path
    """
    if not os.path.isabs(path):
        return
    try:
        with open(path, 'w') as file:
            json.dump(data, file)
    except OSError as e:
        sys.stdout.write(f'[Error] Data could not be saved: {e}')


def reset_temp_folder(path: str):
    '''
    Deletes and re-creates folder housing temporary log files.
    '''
    if os.path.exists(path):
        if os.path.isdir(path):
            shutil.rmtree(path)
        else:
            raise FileExistsError(f'Expected path to folder, got "{path}"')
    os.mkdir(path)


def sanitize_file_name(txt, chr_set='extended') -> str:
    """Converts txt to a valid filename.

    Parameters:
    - :param txt: The path to convert.
    - :param chr_set:
        - 'printable':    Any printable character except those disallowed on Windows/*nix.
        - 'extended':     'printable' + extended ASCII character codes 128-255
        - 'universal':    For almost *any* file system.
    """
    FILLER = '-'
    MAX_LEN = 255  # Maximum length of filename is 255 bytes in Windows and some *nix flavors.

    # Step 1: Remove excluded characters.
    BLACK_LIST = set(chr(127) + r'<>:"/\|?*')
    white_lists = {
        'universal': {'-.0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz'},
        'printable': {chr(x) for x in range(32, 127)} - BLACK_LIST,     # 0-32, 127 are unprintable,
        'extended': {chr(x) for x in range(32, 256)} - BLACK_LIST,
    }
    white_list = white_lists[chr_set]
    result = ''.join(x if x in white_list else FILLER for x in txt)

    # Step 2: Device names, '.', and '..' are invalid filenames in Windows.
    DEVICE_NAMES = (
            'CON', 'PRN', 'AUX', 'NUL', 'COM1', 'COM2', 'COM3', 'COM4', 'COM5', 'COM6', 'COM7',
            'COM8', 'COM9', 'LPT1', 'LPT2', 'LPT3', 'LPT4', 'LPT5', 'LPT6', 'LPT7', 'LPT8', 'LPT9',
            'CONIN$', 'CONOUT$', '..', '.')
    if '.' in txt:
        name, _, ext = result.rpartition('.')
        ext = f'.{ext}'
    else:
        name = result
        ext = ''
    if name in DEVICE_NAMES:
        result = f'-{result}-{ext}'

    # Step 3: Truncate long files while preserving the file extension.
    if len(result) > MAX_LEN:
        result = result[:MAX_LEN - len(ext)] + ext

    # Step 4: Windows does not allow filenames to end with '.' or ' ' or begin with ' '.
    result = re.sub(r"[. ]$", FILLER, result)
    result = re.sub(r"^ ", FILLER, result)

    return result



"""Backend inteface to the OSCR web server"""

import gzip
import json
import os
import tempfile

import OSCR_django_client
from OSCR.utilities import logline_to_str
from OSCR_django_client.api import (CombatlogApi, LadderApi, LadderEntriesApi,
                                    VariantApi)
from PySide6.QtWidgets import QMessageBox

from .datafunctions import CustomThread, analyze_log_callback
from .datamodels import LeagueTableModel, SortingProxy
from .headers import get_ladder_headers
from .style import theme_font
from .subwindows import show_warning, uploadresult_dialog
from .textedit import format_datetime_str

OSCR_SERVER_BACKEND = "https://oscr.stobuilds.com/"
# OSCR_SERVER_BACKEND = "http://127.0.0.1:8000"


def establish_league_connection(self, translate):
    """
    Connects to the league server if not already connected.

    Parameters:
    - :param fetch_ladder: fetches available maps and updates map selector if true
    """
    if self.league_api is None:
        self.league_api = OSCRClient(translate)
        map_fetch_thread = CustomThread(
            self.window, lambda: fetch_and_insert_maps(self)
        )
        map_fetch_thread.start()


def fetch_and_insert_maps(self):
    """
    Retrieves maps from API and inserts them into the list.
    """

    update_default_records(self)
    update_seasonal_records(self)


def update_default_records(self):
    """Update the default records widget"""

    ladders = self.league_api.ladders(variant="Default")
    if ladders is not None:
        self.widgets.ladder_selector.clear()
        for ladder in ladders.results:
            solo = "[Solo]" if ladder.is_solo else ""
            key = f"{ladder.name} ({ladder.difficulty}) {solo}"
            self.league_api.ladder_dict[key] = ladder
            self.widgets.ladder_selector.addItem(key)


def update_seasonal_records(self):
    """Update the seasonal records widget"""

    populate_variants(self)
    ladders = self.league_api.ladders(variant=self.variant_list.currentText())
    if ladders is not None:
        self.widgets.season_ladder_selector.clear()
        for ladder in ladders.results:
            solo = "[Solo]" if ladder.is_solo else ""
            key = f"{ladder.name} ({ladder.difficulty}) {solo}"
            self.league_api.ladder_dict_season[key] = ladder
            self.widgets.season_ladder_selector.addItem(key)


def apply_league_table_filter(self, filter_text: str):
    """
    Sets filter to proxy model of league table
    """
    try:
        self.widgets.ladder_table.model().name_filter = filter_text
    except AttributeError:
        pass


def slot_ladder_default(self, selected_map):
    self.season_selector.clearSelection()
    slot_ladder(self, self.league_api.ladder_dict, selected_map)


def slot_ladder_season(self, selected_map):
    self.map_selector.clearSelection()
    self.favorite_selector.clearSelection()
    slot_ladder(self, self.league_api.ladder_dict_season, selected_map)


def slot_ladder(self, ladder_dict, selected_map):
    """
    Fetches current ladder and puts it into the table.
    """

    if selected_map not in ladder_dict:
        return
    if self.widgets.map_tabber.currentIndex() == 0:
        self.widgets.favorite_ladder_selector.selectionModel().clear()
    else:
        self.widgets.ladder_selector.selectionModel().clear()

    selected_ladder = ladder_dict[selected_map]
    self.league_api.current_ladder_id = selected_ladder.id
    ladder_data = self.league_api.ladder_entries(selected_ladder.id)
    if len(ladder_data.results) >= 50:
        self.league_api.entire_ladder_loaded = False
    self.league_api.pages_loaded = 1
    table_index = list()
    table_data = list()
    logfile_ids = list()

    for entry in ladder_data.results:
        logfile_ids.append(entry.combatlog)
        row = entry.data
        table_index.append(entry.rank)
        table_data.append(
            (
                row["name"],
                row["handle"],
                row["DPS"],
                row["total_damage"],
                row["deaths"],
                row["combat_time"],
                format_datetime_str(entry.var_date),
                row["max_one_hit"],
                row["debuff"],
                row.get("build", "Unknown"),
            )
        )

    model = LeagueTableModel(
        table_data,
        get_ladder_headers(),
        table_index,
        theme_font(self, "table_header"),
        theme_font(self, "table"),
        combatlog_id_list=logfile_ids,
    )
    sorting_proxy = SortingProxy()
    sorting_proxy.setSourceModel(model)
    table = self.widgets.ladder_table
    table.setModel(sorting_proxy)
    table.resizeColumnsToContents()
    # table_header = table.horizontalHeader()
    # for col in range(len(model._header)):
    #     table_header.resizeSection(col, table_header.sectionSize(col) + 5)


def extend_ladder(self):
    """
    Extends the ladder table by 50 newly fetched rows.
    """
    if self.league_api.entire_ladder_loaded:
        return
    if self.league_api.current_ladder_id is None:
        return

    ladder_data = self.league_api.ladder_entries(
        self.league_api.current_ladder_id, self.league_api.pages_loaded + 1
    )
    if ladder_data is not None:
        if len(ladder_data.results) < 50:
            self.league_api.entire_ladder_loaded = True
        self.league_api.pages_loaded += 1
        table_index = list()
        table_data = list()
        logfile_ids = list()
        for entry in ladder_data.results:
            logfile_ids.append(entry.combatlog)
            row = entry.data
            table_index.append(entry.rank)
            table_data.append(
                (
                    row["name"],
                    row["handle"],
                    row["DPS"],
                    row["total_damage"],
                    row["deaths"],
                    row["combat_time"],
                    format_datetime_str(entry.var_date),
                    row["max_one_hit"],
                    row["debuff"],
                    row.get("build", "Unknown"),
                )
            )
        self.widgets.ladder_table.model().sourceModel().extend_data(
            table_index, table_data, logfile_ids
        )


def download_and_view_combat(self, translation):
    """
    Download a combat log and view its contents in the overview / analysis pages.
    """
    table = self.widgets.ladder_table
    table_model = table.model().sourceModel()
    selection = table.selectedIndexes()
    original_index = table.model().mapToSource(selection[0])
    row = original_index.row()
    log_id = table_model._combatlog_id_list[row]
    result = self.league_api.download(log_id)
    result = gzip.decompress(result)
    with tempfile.NamedTemporaryFile(
        mode="w", encoding="utf-8", dir=self.config["templog_folder_path"], delete=False
    ) as file:
        file.write(result.decode())
    analyze_log_callback(
        self, translation, path=file.name, parser_num=1, hidden_path=True
    )
    self.switch_overview_tab(0)
    self.switch_main_tab(0)


def upload_callback(self, translate):
    """
    Helper function to grab the current combat and upload it to the backend.
    """
    if (
        self.parser1.active_combat is None
        or self.parser1.active_combat.log_data is None
    ):
        show_warning(self, "OSCR - Logfile Upload", self._("No data to upload."))
        return

    establish_league_connection(self, translate)

    with tempfile.NamedTemporaryFile(delete=False) as file:
        data = gzip.compress(
            "".join(
                [logline_to_str(line) for line in self.parser1.active_combat.log_data]
            ).encode()
        )
        file.write(data)
        file.flush()
    res = self.league_api.upload(file.name)
    if res:
        uploadresult_dialog(self, res, translate)
    os.remove(file.name)


def populate_variants(self):
    """Populate the list of variants"""

    # Only populate the table once.
    if self.variant_list.count():
        return

    variants = self.league_api.variants(ordering="-start_date")
    for variant in variants.results:
        if variant.name != "Default":
            self.variant_list.addItem(variant.name)

    self.variant_list.setCurrentIndex(0)


class OSCRClient:
    def __init__(self, translate):
        """Initialize an instance of the OSCR backlend client"""

        self.address = OSCR_SERVER_BACKEND
        self._ = translate

        self.api_client = OSCR_django_client.api_client.ApiClient()
        self.api_client.configuration.host = self.address
        self.api_combatlog = CombatlogApi(api_client=self.api_client)
        self.api_ladder = LadderApi(api_client=self.api_client)
        self.api_ladder_entries = LadderEntriesApi(api_client=self.api_client)
        self.api_variant = VariantApi(api_client=self.api_client)
        self.ladder_dict: dict = dict()
        self.ladder_dict_season: dict = dict()
        self.current_ladder_id = None
        self.pages_loaded: int = -1
        self.entire_ladder_loaded: bool = True

    def upload(self, filename):
        """Upload a combat log located at path for analysis"""

        try:
            return self.api_combatlog.combatlog_uploadv2(file=filename)
        except OSCR_django_client.exceptions.ServiceException as e:
            reply = QMessageBox()
            reply.setWindowTitle("Open Source Combatlog Reader")
            try:
                data = json.loads(e.body)
                reply.setText(
                    data.get("detail", self._("Failed to parse error from server"))
                )
            except Exception as e:
                reply.setText(self._("Failed to parse error from server"))
            reply.exec()

    def download(self, id):
        """Download a combat log"""
        try:
            return self.api_combatlog.combatlog_download(id=id)
        except OSCR_django_client.exceptions.ServiceException as e:
            reply = QMessageBox()
            reply.setWindowTitle("Open Source Combatlog Reader")
            try:
                data = json.loads(e.body)
                reply.setText(
                    data.get("detail", self._("Failed to parse error from server"))
                )
            except Exception as e:
                reply.setText(self._("Failed to parse error from server"))
            reply.exec()

        return None

    def ladders(self, **kwargs):
        """Fetch the list of ladders"""
        try:
            return self.api_ladder.ladder_list(**kwargs)
        except OSCR_django_client.exceptions.ServiceException as e:
            reply = QMessageBox()
            reply.setWindowTitle("Open Source Combatlog Reader")
            try:
                data = json.loads(e.body)
                reply.setText(
                    data.get("detail", self._("Failed to parse error from server"))
                )
            except Exception as e:
                reply.setText(self._("Failed to parse error from server"))
            reply.exec()

        return None

    def ladder_entries(self, id, page=1):
        """Fetch the nth page of ladder entries"""
        try:
            return self.api_ladder_entries.ladder_entries_list(
                ladder=str(id),
                page=page,
                ordering="-data__DPS",
                page_size=50,
            )
        except OSCR_django_client.exceptions.ServiceException as e:
            reply = QMessageBox()
            reply.setWindowTitle("Open Source Combatlog Reader")
            try:
                data = json.loads(e.body)
                reply.setText(
                    data.get("detail", self._("Failed to parse error from server"))
                )
            except Exception as e:
                reply.setText(self._("Failed to parse error from server"))
            reply.exec()

        return None

    def variants(self, **kwargs):
        """Return a list of Variants"""

        try:
            return self.api_variant.variant_list(**kwargs)
        except OSCR_django_client.exceptions.ServiceException as e:
            reply = QMessageBox()
            reply.setWindowTitle("Open Source Combatlog Reader")
            try:
                data = json.loads(e.body)
                reply.setText(
                    data.get("detail", self._("Failed to parse error from server"))
                )
            except Exception as e:
                reply.setText(self._("Failed to parse error from server"))
            reply.exec()

        return None



import copy

from PySide6.QtGui import QFont

WEIGHT_CONVERSION = {
    'normal': QFont.Weight.Normal,
    'bold': QFont.Weight.Bold,
    'extrabold': QFont.Weight.ExtraBold,
    'medium': QFont.Weight.Medium
}


def get_style(self, widget, override: dict = {}) -> str:
    """
    Returns style sheet according to default style of widget with override style.

    Parameters:
    - :param widget: None or str -> name of the widget style in self.theme (may be empty or None if
    only the style in override should be applied)
    - :param override: dict -> contains additional style (optional)

    :return: str containing css style sheet
    """
    if widget is None or widget == '':
        return get_css(self, override)
    elif widget != 'app' and widget != 'defaults' and widget != 's.c' and widget in self.theme:
        if len(override) > 0:
            style = merge_style(self, self.theme[widget], override)
        else:
            style = self.theme[widget]
        return get_css(self, style)


def get_style_class(self, class_name: str, widget, override={}) -> str:
    """
    Returns style sheet according to default style of widget with override style. Style only
    applies to class_name. Sub-controls, pseudo-states and descendant selectors (marked with "~")
    defined in self.theme and override are correctly handled.

    Parameters:
    - :param class_name: str -> name of the widget to be styled
    - :param widget: None or str -> name of the widget style in self.theme (may be empty or None if
    only the style in override should be applied)
    - :param override: dict -> contains additional style (optional)

    :return: str containing css style sheet
    """
    if widget is None or widget == '':
        style = override
    elif widget != 'app' and widget != 'defaults' and widget != 's.c' and widget in self.theme:
        if len(override) > 0:
            style = merge_style(self, self.theme[widget], override)
        else:
            style = self.theme[widget]
    else:
        raise KeyError(
                f'Parameter widget=`{widget}` must be None or key of self.theme '
                'except `app` or `defaults`.')
    main = f'{class_name} {{{get_css(self, style)}}}'
    for k, v in style.items():
        if k.startswith(':'):
            main += f''' {class_name}{k} {{{get_css(self, v)}}}'''
        elif k.startswith('~'):
            main += f' {get_style_class(self, f"{class_name} {k[1:]}", None, v)}'
    return main


def merge_style(self, s1: dict, s2: dict) -> dict:
    """
    Returns new dictionary where the given styles are merged.
    Up to one sub-dictionary is merged recursively.

    Parameters:
    - :param s1: Style-dict 1
    - :param s2: Style-dict 2

    :return: merged dictionary
    """
    result = copy.deepcopy(s1)
    for k, v in s2.items():
        if k in result.keys() and isinstance(result[k], dict) and isinstance(v, dict):
            result[k].update(v)
            continue
        result[k] = v
    return result


def get_css(self, style: dict) -> str:
    """
    Converts style dictionary into css style sheet. Escapes '@' - shortcuts with their respective
    values.
    """
    css = str()
    ui_scale = self.config['ui_scale']
    for key, val in style.items():
        if isinstance(val, str) and val.startswith('@'):
            v = self.theme['defaults'][val[1:]]
        else:
            v = val
        if key.startswith(':') or key.startswith('~') or key == 'font':
            continue
        elif isinstance(v, int):
            css += f'{key}:{v * ui_scale}px;'
        elif isinstance(v, tuple):
            css += f'''{key}:{'px '.join(map(lambda s: str(s * ui_scale), v))}px;'''
        else:
            css += f'{key}:{v};'
    return css


def theme_font(self, key, font_spec=()) -> QFont:
    """
    Returns QFont object with font specified in self.theme or font_spec. Adds default fallback font
    families.

    Parameters:
    - :param key: key in self.theme to access font tuple like: self.theme[key]['font']
    - :param font_spec: font tuple consisting of family, size and weight OR font shortcut (optional)

    :return: configured QFont object
    """
    try:
        if len(font_spec) != 3 and isinstance(font_spec, tuple):
            font_spec = self.theme[key]['font']
        if isinstance(font_spec, str) and font_spec.startswith('@'):
            font = self.theme['defaults'][font_spec[1:]]
        else:
            font = font_spec
    except KeyError:
        font = self.theme['app']['font']
    font_family = (font[0], *self.theme['app']['font-fallback'])
    font_size = int(font[1] * self.config['ui_scale'])
    try:
        font_weight = WEIGHT_CONVERSION[font[2]]
    except KeyError:
        font_weight = QFont.Weight.Normal
    return QFont(font_family, font_size, font_weight)


def create_style_sheet(self, d: dict) -> str:
    """
    Creates Stylesheet from dictionary. Dictionary keys represent css selector.

    Parameters:
    - :param d: dict -> style dictionary

    :return: string containing css sheet
    """
    style = str()
    for s, v in d.items():
        style += f'{s} {{{get_css(self, v)}}}'
    return style



import os
import json

from OSCR import OSCR
from PySide6.QtCore import Qt, QThread, Signal
from PySide6.QtWidgets import QMessageBox

from .callbacks import switch_main_tab, switch_overview_tab, trim_logfile
from .datamodels import DamageTreeModel, HealTreeModel, TreeSelectionModel
from .displayer import create_overview
from .headers import get_heal_tree_headers, get_tree_headers
from .subwindows import log_size_warning, show_warning, split_dialog
from .textedit import format_damage_number, format_damage_tree_data, format_heal_tree_data


class CustomThread(QThread):
    """
    Subclass of QThread able to execute an arbitrary function in a seperate thread.
    """
    result = Signal(tuple)

    def __init__(self, parent, func) -> None:
        self._func = func
        super().__init__(parent)

    def run(self):
        r = self._func()
        self.result.emit((r,))


def init_parser(self):
    """
    Initializes Parser.
    """
    self.parser1 = OSCR(settings=self.parser_settings)
    # self.parser2 = OSCR()


def analyze_log_callback(
        self, translate, combat_id=None, path=None, parser_num: int = 1, hidden_path=False):
    """
    Wrapper function for retrieving and showing data. Callback of "Analyse" and "Refresh" button.

    Parameters:
    - :param combat_id: id of older combat (0 -> latest combat in the file;
    len(...) - 1 -> oldest combat)
    - :param path: path to combat log file
    - :param parser_num: 1 or 2; to select self.parser1 or self.parser2
    - :param hidden_path: True when settings should not be updated with log path
    """
    if combat_id == -1 or combat_id == self.current_combat_id:
        return
    if parser_num == 1:
        parser: OSCR = self.parser1
    elif parser_num == 2:
        parser: OSCR = self.parser2
    else:
        return

    if self._ is None:
        self._ = translate

    # initial run / click on the Analyze buttonQGuiApplication
    if combat_id is None:
        if not path or not os.path.isfile(path):
            show_warning(
                    self, self._('Invalid Logfile'),
                    self._('The Logfile you are trying to open does not exist.'))
            return
        if not hidden_path and path != self.settings.value('log_path'):
            self.settings.setValue('log_path', path)

        parser.log_path = path
        try:
            parser.analyze_log_file()
        except FileExistsError:
            if self.settings.value('log_size_warning', type=bool):
                action = log_size_warning(self, translate)
                if action == 'split dialog':
                    split_dialog(self)
                    return
                elif action == 'trim':
                    trim_logfile(self)
                    parser.analyze_log_file()
                elif action == 'continue':
                    parser.analyze_massive_log_file()
                elif action == 'cancel':
                    return

        except Exception as ex:
            error = QMessageBox()
            error.setWindowTitle("Open Source Combatlog Reader")
            try:
                print(ex)
                error_message = str(ex)[:60]
                message = (
                    f"{self._('Failed to analyze the log file.')}\n\n"
                    f"{self._('Reason:')}\n{error_message}\n\n"
                    f"{self._('Please report this issue to Discord OSCR-Support channel.')}"
                )
                error.setText(message)
            except Exception as ex:
                print(ex)
                error_message = str(ex)[:60]
                message = (
                    f"{self._('Failed to analyze the log file.')}\n\n"
                    f"{self._('Reason:')}\n{error_message}\n\n"
                    f"{self._('Please report this issue to Discord OSCR-Support channel.')}"
                )
                error.setText(message)

            error.setWindowTitle(self._("Open Source Combatlog Reader"))
            error.setIcon(QMessageBox.Critical)
            error.exec()

        self.current_combats.clear()
        self.current_combats.addItems(parser.analyzed_combats)
        self.current_combats.setCurrentRow(0)
        self.current_combat_id = 0
        self.current_combat_path = path
        self.widgets.navigate_up_button.setEnabled(parser.navigation_up)
        self.widgets.navigate_down_button.setEnabled(parser.navigation_down)

        analysis_thread = CustomThread(self.window, lambda: parser.full_combat_analysis(0))
        analysis_thread.result.connect(lambda result: analysis_data_slot(self, result))
        analysis_thread.start(QThread.Priority.IdlePriority)

    # subsequent run / click on older combat
    elif isinstance(combat_id, int):
        self.current_combat_id = combat_id
        analysis_thread = CustomThread(self.window, lambda: parser.full_combat_analysis(combat_id))
        analysis_thread.result.connect(lambda result: analysis_data_slot(self, result))
        analysis_thread.start(QThread.Priority.IdlePriority)

    # reset tabber
    switch_main_tab(self, 0)
    switch_overview_tab(self, self.settings.value('first_overview_tab', type=int))


def copy_summary_callback(self, translate, parser_num: int = 1):
    """
    Callback to set the combat summary of the active combat to the user's clippboard

    Parameters:
    - :param parser_num: which parser to take the data from
    """
    if parser_num == 2:
        parser = self.parser2
    else:
        parser = self.parser1

    if not parser.active_combat:
        return

    if self._ is None:
        self._ = translate

    duration = self.parser1.active_combat.duration.total_seconds()
    combat_time = f'{int(duration / 60):02}:{duration % 60:02.0f}'

    summary = f'OSCR {parser.active_combat.map}'
    difficulty = parser.active_combat.difficulty
    if difficulty and isinstance(difficulty, str) and difficulty != 'Unknown':
        summary += f' ({difficulty}) - DPS [{combat_time}]: '
    else:
        summary += f' - DPS [{combat_time}]: '
    players = sorted(
        self.parser1.active_combat.player_dict.values(),
        reverse=True,
        key=lambda player: player.DPS,
    )
    parts = list()
    for player in players:
        parts.append(
                f"{player.handle} {player.DPS:,.0f}")
    summary += " | ".join(parts)

    self.app.clipboard().setText(summary)


def analysis_data_slot(self, item_tuple: tuple):
    """
    Inserts the data retrieved from the parser into the respective tables

    Parameters:
    - :param item_tuple: tuple containing only the root item of the data model
    """
    create_overview(self)
    populate_analysis(self, *item_tuple)
    self.widgets.main_menu_buttons[1].setDisabled(False)


def populate_analysis(self, root_items: tuple):
    """
    Populates the Analysis' treeview table.
    """
    damage_out_item, damage_in_item, heal_out_item, heal_in_item = root_items

    damage_out_table = self.widgets.analysis_table_dout
    damage_out_model = DamageTreeModel(
            damage_out_item, self.theme_font('tree_table_header'), self.theme_font('tree_table'),
            self.theme_font('', self.theme['tree_table']['::item']['font']), get_tree_headers())
    damage_out_table.setModel(damage_out_model)
    damage_out_root_index = damage_out_model.createIndex(0, 0, damage_out_model._root)
    damage_out_table.expand(damage_out_model.index(0, 0, damage_out_root_index))
    damage_out_table.sortByColumn(1, Qt.SortOrder.AscendingOrder)
    damage_out_table.setSelectionModel(TreeSelectionModel(damage_out_model))

    damage_in_table = self.widgets.analysis_table_dtaken
    damage_in_model = DamageTreeModel(
            damage_in_item, self.theme_font('tree_table_header'), self.theme_font('tree_table'),
            self.theme_font('', self.theme['tree_table']['::item']['font']), get_tree_headers())
    damage_in_table.setModel(damage_in_model)
    damage_in_root_index = damage_in_model.createIndex(0, 0, damage_in_model._root)
    damage_in_table.expand(damage_in_model.index(0, 0, damage_in_root_index))
    damage_in_table.sortByColumn(1, Qt.SortOrder.AscendingOrder)
    damage_in_table.setSelectionModel(TreeSelectionModel(damage_in_model))

    heal_out_table = self.widgets.analysis_table_hout
    heal_out_model = HealTreeModel(
            heal_out_item, self.theme_font('tree_table_header'), self.theme_font('tree_table'),
            self.theme_font('', self.theme['tree_table']['::item']['font']),
            get_heal_tree_headers())
    heal_out_table.setModel(heal_out_model)
    heal_out_root_index = damage_in_model.createIndex(0, 0, heal_out_model._root)
    heal_out_table.expand(heal_out_model.index(0, 0, heal_out_root_index))
    heal_out_table.sortByColumn(1, Qt.SortOrder.AscendingOrder)
    heal_out_table.setSelectionModel(TreeSelectionModel(heal_out_model))

    heal_in_table = self.widgets.analysis_table_hin
    heal_in_model = HealTreeModel(
            heal_in_item, self.theme_font('tree_table_header'), self.theme_font('tree_table'),
            self.theme_font('', self.theme['tree_table']['::item']['font']),
            get_heal_tree_headers())
    heal_in_table.setModel(heal_in_model)
    heal_in_root_index = damage_in_model.createIndex(0, 0, heal_in_model._root)
    heal_in_table.expand(heal_in_model.index(0, 0, heal_in_root_index))
    heal_in_table.sortByColumn(1, Qt.SortOrder.AscendingOrder)
    heal_in_table.setSelectionModel(TreeSelectionModel(heal_in_model))

    update_shown_columns_dmg(self)
    update_shown_columns_heal(self)


def update_shown_columns_dmg(self):
    """
    Hides / shows columns of the dmg analysis tables.
    """
    dout_table = self.widgets.analysis_table_dout
    dtaken_table = self.widgets.analysis_table_dtaken
    for i in range(self.settings.value('dmg_columns_length', type=int)):
        state = self.settings.value(f'dmg_columns|{i}', type=bool)
        if state:
            dout_table.showColumn(i + 1)
            dtaken_table.showColumn(i + 1)
        else:
            dout_table.hideColumn(i + 1)
            dtaken_table.hideColumn(i + 1)


def update_shown_columns_heal(self):
    """
    Hides / shows columns of the heal analysis tables.
    """
    hout_table = self.widgets.analysis_table_hout
    hin_table = self.widgets.analysis_table_hin
    for i in range(self.settings.value('heal_columns_length', type=int)):
        state = self.settings.value(f'heal_columns|{i}', type=bool)
        if state:
            hout_table.showColumn(i + 1)
            hin_table.showColumn(i + 1)
        else:
            hout_table.hideColumn(i + 1)
            hin_table.hideColumn(i + 1)


def resize_tree_table(tree):
    """
    Resizes the columns of the given tree table to fit its contents.

    Parameters:
    - :param tree: QTreeView -> tree to be resized
    """
    for col in range(tree.header().count()):
        width = max(tree.sizeHintForColumn(col), tree.header().sectionSizeHint(col)) + 5
        tree.header().resizeSection(col, width)


def copy_analysis_table_callback(self):
    """
    Copies the current selection of analysis table as tab-delimited table
    """
    print('Table COPY')
    if self.widgets.main_tabber.currentIndex() != 1:
        return
    current_tab = self.widgets.analysis_tabber.currentIndex()
    current_table = self.widgets.analysis_table[current_tab]
    if current_tab <= 1:
        format_function = format_damage_tree_data
    else:
        format_function = format_heal_tree_data
    selection: list = current_table.selectedIndexes()
    if selection:
        selection.sort(key=lambda index: (index.row(), index.column()))
        output = list()
        last_row = -1
        for cell_index in selection:
            col = cell_index.column()
            if cell_index.row() != last_row:
                output.append(list())
            output[-1].append(cell_index.internalPointer().get_data(col))
            last_row = cell_index.row()
        print(output)
        output_text = '\n'.join(map(lambda row: '\t'.join(map(str, row)), output))
        print(output_text)
        self.app.clipboard().setText(output_text)


def copy_analysis_callback(self):
    """
    Callback for copy button on analysis tab
    """
    current_tab = self.widgets.analysis_tabber.currentIndex()
    current_table = self.widgets.analysis_table[current_tab]
    copy_mode = self.widgets.analysis_copy_combobox.currentText()
    if copy_mode == self._('Selection'):
        if current_tab <= 1:
            current_header = get_tree_headers()
            format_function = format_damage_tree_data
        else:
            current_header = get_heal_tree_headers()
            format_function = format_heal_tree_data
        selection = current_table.selectedIndexes()
        if selection:
            selection_dict = dict()
            for selected_cell in selection:
                column = selected_cell.column()
                row_name = selected_cell.internalPointer().get_data(0)
                if row_name not in selection_dict:
                    selection_dict[row_name] = dict()
                if column != 0:
                    cell_data = selected_cell.internalPointer().get_data(column)
                    selection_dict[row_name][column] = cell_data
            output = ['{ OSCR }']
            for row_name, row_data in selection_dict.items():
                formatted_row = list()
                for col, value in row_data.items():
                    formatted_row.append(f'[{current_header[col]}] {format_function(value, col)}')
                formatted_row_name = ''.join(row_name) if isinstance(row_name, tuple) else row_name
                output.append(f"{formatted_row_name}: {' | '.join(formatted_row)}")
            output_string = '\n'.join(output)
            self.app.clipboard().setText(output_string)
    elif copy_mode == self._('Global Max One Hit'):
        if current_tab <= 1:
            max_one_hit_col = 4
            prefix = self._('Max One Hit')
        else:
            max_one_hit_col = 7
            prefix = self._('Max One Heal')
        max_one_hits = []
        for player_item in current_table.model()._player._children:
            max_one_hits.append((player_item.get_data(max_one_hit_col), player_item))
        max_one_hit, max_one_hit_item = max(max_one_hits, key=lambda x: x[0])
        max_one_hit_ability = max(
                max_one_hit_item._children, key=lambda x: x.get_data(max_one_hit_col))
        max_one_hit_ability = max_one_hit_ability.get_data(0)
        if isinstance(max_one_hit_ability, tuple):
            max_one_hit_ability = ''.join(max_one_hit_ability)
        output_string = (f'OSCR {prefix}: {max_one_hit:,.2f} '
                         f'(`{"".join(max_one_hit_item.get_data(0))}` – '
                         f'{max_one_hit_ability})')
        self.app.clipboard().setText(output_string)
    elif copy_mode == self._('Max One Hit'):
        if current_tab <= 1:
            max_one_hit_col = 4
            prefix = self._('Max One Hit')
        else:
            max_one_hit_col = 7
            prefix = self._('Max One Heal')
        selection = current_table.selectedIndexes()
        if selection:
            selected_row = selection[0].internalPointer()
            if selected_row._children:
                max_one_hit_item = max(
                        selected_row._children, key=lambda child: child.get_data(max_one_hit_col))
                max_one_hit = max_one_hit_item.get_data(max_one_hit_col)
                max_one_hit_ability = max_one_hit_item.get_data(0)
                if isinstance(max_one_hit_ability, tuple):
                    max_one_hit_ability = ''.join(max_one_hit_ability)
                output_string = (f'OSCR {prefix}: {max_one_hit:,.2f} '
                                 f'(`{"".join(selected_row.get_data(0))}` – '
                                 f'{max_one_hit_ability})')
                self.app.clipboard().setText(output_string)
    elif copy_mode == self._('Magnitude'):
        if current_tab == 0:
            prefix = self._('Total Damage Out')
        elif current_tab == 1:
            prefix = self._('Total Damage Taken')
        elif current_tab == 2:
            prefix = self._('Total Heal Out')
        else:
            prefix = self._('Total Heal In')
        magnitudes = list()
        for player_item in current_table.model()._player._children:
            magnitudes.append((player_item.get_data(2), ''.join(player_item.get_data(0))))
        magnitudes.sort(key=lambda x: x[0], reverse=True)
        magnitudes = [f"`[{''.join(player)}]` {magnitude:,.2f}" for magnitude, player in magnitudes]
        output_string = (f'OSCR {prefix}: {" | ".join(magnitudes)}')
        self.app.clipboard().setText(output_string)
    elif copy_mode == self._('Magnitude / s'):
        if current_tab == 0:
            prefix = self._('Total DPS Out')
        elif current_tab == 1:
            prefix = self._('Total DPS Taken')
        elif current_tab == 2:
            prefix = self._('Total HPS Out')
        else:
            prefix = self._('Total HPS In')
        magnitudes = list()
        for player_item in current_table.model()._player._children:
            magnitudes.append((player_item.get_data(1), ''.join(player_item.get_data(0))))
        magnitudes.sort(key=lambda x: x[0], reverse=True)
        magnitudes = [f"`[{''.join(player)}]` {magnitude:,.2f}" for magnitude, player in magnitudes]
        output_string = (f'OSCR {prefix}: {" | ".join(magnitudes)}')
        self.app.clipboard().setText(output_string)



from typing import Iterable
import sys

from OSCR import TreeItem
from PySide6.QtCore import QAbstractItemModel, QAbstractTableModel, QItemSelectionModel
from PySide6.QtCore import QItemSelection, QModelIndex, QSortFilterProxyModel, Qt
from PySide6.QtGui import QColor, QFont

ARIGHT = Qt.AlignmentFlag.AlignRight
ALEFT = Qt.AlignmentFlag.AlignLeft
ACENTER = Qt.AlignmentFlag.AlignCenter
AVCENTER = Qt.AlignmentFlag.AlignVCenter


class TableModel(QAbstractTableModel):
    def __init__(
            self, data, header: Iterable, index: Iterable, header_font: QFont, cell_font: QFont):
        """
        Creates table model from supplied data.

        Parameters:
        - :param data: data to be displayed without index or header; two-dimensional iterable;
        must support .extend() function
        - :param header: column headings
        - :param index: row index
        - :param header_font: font to style the column headings with
        - :param cell_font: font to style the cells with
        """
        super().__init__()
        self._data = data
        self._header = tuple(header)
        self._index = list(index)
        self._header_font = header_font
        self._cell_font = cell_font

    def rowCount(self, index):
        return len(self._data)

    def columnCount(self, index):
        try:
            return len(self._data[0])  # all columns must have the same length
        except IndexError:
            return 0

    def headerData(self, section, orientation, role):
        # section is the index of the column/row.
        if role == Qt.ItemDataRole.DisplayRole:
            if orientation == Qt.Orientation.Horizontal:
                return self._header[section]

            if orientation == Qt.Orientation.Vertical:
                return self._index[section]

        if role == Qt.ItemDataRole.FontRole:
            return self._header_font

        if role == Qt.ItemDataRole.TextAlignmentRole:
            if orientation == Qt.Orientation.Horizontal:
                return ACENTER

            if orientation == Qt.Orientation.Vertical:
                return AVCENTER + ARIGHT


class OverviewTableModel(TableModel):
    """
    Model for overview table
    """
    def data(self, index, role):
        if role == Qt.ItemDataRole.DisplayRole:
            current_col = index.column()
            cell = self._data[index.row()][current_col]
            column = index.column()
            if column == 0:
                return f'{cell:.1f}s'
            elif column in (1, 2, 7, 10, 13, 14, 15):
                return f'{cell:,.2f}'
            elif column in (3, 4, 5, 6, 8, 11, 12):
                return f'{cell:,.2f}%'
            elif column in (9, 16, 17, 18, 19, 20, 21, 22):
                return str(cell)
            return cell

        if role == Qt.ItemDataRole.FontRole:
            return self._cell_font

        if role == Qt.ItemDataRole.TextAlignmentRole:
            return AVCENTER + ARIGHT


class LeagueTableModel(TableModel):
    """
    Model for league table
    """
    def __init__(self, *ar, combatlog_id_list: list = [], **kw):
        super().__init__(*ar, **kw)
        self._combatlog_id_list = combatlog_id_list

    def data(self, index, role):
        if role == Qt.ItemDataRole.DisplayRole:
            current_col = index.column()
            cell = self._data[index.row()][current_col]
            column = index.column()
            if column == 5:
                return f'{cell:.1f}s'
            elif column in (2, 3, 7):
                return f'{cell:,.2f}'
            elif column == 8:
                return f'{cell:,.2f}%'
            elif column == 4:
                return str(cell)
            return cell

        if role == Qt.ItemDataRole.FontRole:
            return self._cell_font

        if role == Qt.ItemDataRole.TextAlignmentRole:
            if index.column() == 1:
                return AVCENTER + ALEFT
            return AVCENTER + ARIGHT

    def headerData(self, section, orientation, role):
        if role == Qt.ItemDataRole.FontRole and orientation == Qt.Orientation.Vertical:
            return self._cell_font
        return super().headerData(section, orientation, role)

    def extend_data(self, index: list, rows: list, combatlog_ids: list):
        current_row_count = len(self._index)
        self.beginInsertRows(QModelIndex(), current_row_count, current_row_count + len(index) - 1)
        self._index.extend(index)
        self._data.extend(rows)
        self._combatlog_id_list.extend(combatlog_ids)
        self.endInsertRows()


class LiveParserTableModel(TableModel):
    """
    Model for LiveParser Table
    """
    def __init__(self, *args, legend_col=None, colors=None, **kwargs):
        super().__init__(*args, **kwargs)
        self._legend_column = legend_col
        if colors is not None:
            self._colors = [QColor.fromString(color) for color in colors]
        else:
            self._colors = None

    def data(self, index, role):
        if role == Qt.ItemDataRole.DisplayRole:
            column = index.column()
            data = self._data[index.row()][column]
            if column in (0, 4):
                return f'{data:,.2f}'
            elif column == 1:
                return f'{data:.1f}s'
            elif column == 2:
                if data == 0:
                    return '---.--%'
                return f'{data:,.2f}%'
            elif column == 3:
                return f'{data:,.2f}%'
            return str(data)

        if role == Qt.ItemDataRole.FontRole:
            return self._cell_font

        if role == Qt.ItemDataRole.TextAlignmentRole:
            return AVCENTER + ARIGHT

        if role == Qt.ItemDataRole.DecorationRole:
            if self._legend_column is not None and index.column() == self._legend_column:
                row = index.row()
                if row < len(self._colors):
                    return self._colors[row]
            return None

    def headerData(self, section, orientation, role):
        # section is the index of the column/row.
        if role == Qt.ItemDataRole.DisplayRole:
            if orientation == Qt.Orientation.Horizontal:
                return self._header[section]

            if orientation == Qt.Orientation.Vertical:
                try:
                    return self._index[section]
                except IndexError:
                    sys.stdout.write(f'Section:{section}|Data{self._data}|Index{self._index}\n')

        if role == Qt.ItemDataRole.FontRole:
            return self._header_font

        if role == Qt.ItemDataRole.TextAlignmentRole:
            if orientation == Qt.Orientation.Horizontal:
                return ACENTER

            if orientation == Qt.Orientation.Vertical:
                return AVCENTER + ARIGHT

    def replace_data(self, index: list, rows: list):
        self.beginResetModel()
        self._index = index
        self._data = rows
        self.endResetModel()


class SortingProxy(QSortFilterProxyModel):
    def __init__(self):
        super().__init__()
        self._name_filter: str = ''

    @property
    def name_filter(self) -> str:
        return self._name_filter

    @name_filter.setter
    def name_filter(self, filter_value: str):
        self._name_filter = filter_value
        self.invalidateFilter()

    def lessThan(self, left, right):
        links = self.sourceModel()._data[left.row()][left.column()]
        rechts = self.sourceModel()._data[right.row()][right.column()]
        return links > rechts  # inverted operator to make descending sort come up first

    def filterAcceptsRow(self, source_row: int, source_parent: QModelIndex) -> bool:
        filter_value = self.name_filter
        if filter_value:
            full_name = ''.join(self.sourceModel()._data[source_row][0:2]).casefold()
            return filter_value.casefold() in full_name
        return True


class TreeModel(QAbstractItemModel):
    """
    Data model for the analysis table
    """
    def __init__(
            self, root_item, header_font: QFont, name_font: QFont, cell_font: QFont,
            header_data=None):
        """
        Initializes Tree Model with data in root item.

        Parameters:
        - :param root_item: item supporting the following operations:
            - function "get_child(row: int)" returning the n-th child of the item; None if not
            exists
            - function "get_data(column: int)" returning the data of the item at the given column;
            None if
            not exists
            - function "append_child(item)" adds the given item as child to the item
            - property "parent" containing the parent item of the item; None for the root item
            - property "row" containing the row number which the item is stored at in its parent
            item
            - property "child_count" containing the number of children the item has
            - property "column_count" containing the number of columns the items data row
        - :param header_data: the headers using in the analysis UI.

        The item may already contain childen.
        - :param header_font: font used for the header
        - :param name_font: font used for the first column
        - :param cell_font: font used for the second to last column
        """
        super().__init__()
        self._root = root_item
        self._root_index = self.createIndex(0, 0, self._root)
        self._header_font = header_font
        self._name_font = name_font
        self._cell_font = cell_font
        if header_data is None:
            self._header_data = self._root.data
        else:
            self._header_data = header_data
        if root_item.get_child(0).get_data(0) == 'Player':
            self._player = root_item.get_child(0)
            self._npc = root_item.get_child(1)
        else:
            self._player = root_item.get_child(1)
            self._npc = root_item.get_child(0)

    def sort(self, column: int, order: Qt.SortOrder):
        if order == Qt.SortOrder.AscendingOrder:
            descending = True
        else:
            descending = False
        self.layoutAboutToBeChanged.emit()
        self.recursive_sort(self._root._children[0], column, descending)
        self.recursive_sort(self._root._children[1], column, descending)
        self.layoutChanged.emit()

    def recursive_sort(self, item: TreeItem, column: int, desc):
        if item.child_count > 0:
            for child in item._children:
                self.recursive_sort(child, column, desc)
            item._children.sort(key=lambda row: row.get_data(column), reverse=desc)

    def index(self, row: int, column: int, parent: QModelIndex) -> QModelIndex | None:
        if not self.hasIndex(row, column, parent):
            return QModelIndex()
        if not parent.isValid():
            p = self._root
        else:
            p = parent.internalPointer()
        c = p.get_child(row)
        if c is not None:
            return self.createIndex(row, column, c)
        return QModelIndex()

    def parent(self, index: QModelIndex) -> QModelIndex | None:
        if not index.isValid():
            return QModelIndex()
        child = index.internalPointer()
        parent = child.parent
        if parent is None:
            return QModelIndex()
        return self.createIndex(parent.row, 0, parent)

    def rowCount(self, parent: QModelIndex) -> int:
        if not parent.isValid():
            parent = self._root
        else:
            parent = parent.internalPointer()
        return parent.child_count

    def columnCount(self, parent: QModelIndex) -> int:
        if parent.isValid():
            return parent.internalPointer().column_count
        return self._root.column_count

    def flags(self, index: QModelIndex) -> Qt.ItemFlag:
        if not index.isValid():
            return Qt.ItemFlag.NoItemFlags
        return super().flags(index)

    def headerData(self, section, orientation, role) -> str:
        if role == Qt.ItemDataRole.DisplayRole:
            return self._header_data[section]
        elif role == Qt.ItemDataRole.FontRole:
            return self._header_font
        elif role == Qt.ItemDataRole.TextAlignmentRole:
            return ACENTER
        return None


class DamageTreeModel(TreeModel):
    """
    Tree Model subclass for the damage tables
    """
    def data(self, index: QModelIndex, role: int) -> str:
        if not index.isValid():
            return ''
        column = index.column()
        if role == Qt.ItemDataRole.DisplayRole:
            data = index.internalPointer().get_data(column)
            if data == '':
                return ''
            if column == 0:
                if isinstance(data, tuple):
                    return ''.join(data)
                return data
            elif column in (3, 5, 6, 7):
                return f'{data * 100:,.2f}%'
            elif column in (1, 2, 4, 13, 14, 15, 16, 17, 18):
                return f'{data:,.2f}'
            elif column in (8, 9, 10, 11, 12, 20, 21):
                return f'{data:,.0f}'
            elif column == 19:
                return f'{data}s'
        elif role == Qt.ItemDataRole.FontRole:
            if column == 0:
                return self._name_font
            return self._cell_font
        elif role == Qt.ItemDataRole.TextAlignmentRole:
            if column != 0:
                return AVCENTER + ARIGHT
            else:
                return AVCENTER + ALEFT
        elif role == -13:
            return index.internalPointer().get_data(column)
        return None


class HealTreeModel(TreeModel):
    """
    Tree Model subclass for the heal tables
    """
    def data(self, index: QModelIndex, role: int) -> str:
        if not index.isValid():
            return ''
        column = index.column()
        if role == Qt.ItemDataRole.DisplayRole:
            data = index.internalPointer().get_data(column)
            if data == '':
                return ''
            if column == 0:
                if isinstance(data, tuple):
                    return ''.join(data)
                return data
            elif column == 8:
                return f'{data * 100:,.2f}%'
            elif column in (1, 2, 3, 4, 5, 6, 7, 17, 18):
                return f'{data:,.2f}'
            elif column in (9, 10, 12, 13):
                return f'{data:,.0f}'
            elif column == 11:
                return f'{data}s'
        elif role == Qt.ItemDataRole.FontRole:
            if column == 0:
                return self._name_font
            return self._cell_font
        elif role == Qt.ItemDataRole.TextAlignmentRole:
            if column != 0:
                return AVCENTER + ARIGHT
            else:
                return AVCENTER + ALEFT
        elif role == -13:
            return index.internalPointer().get_data(column)
        return None


class TreeSelectionModel(QItemSelectionModel):
    """
    Implements custom selection behavior for analysis tables.
    """
    def __init__(self, model: QAbstractItemModel):
        super().__init__(model)

    def select(
            self, index_or_selection: QModelIndex | QItemSelection,
            flag: QItemSelectionModel.SelectionFlag):
        if isinstance(index_or_selection, QItemSelection):
            try:
                if index_or_selection.indexes()[0].column() == 0:
                    super().select(
                        index_or_selection, flag | QItemSelectionModel.SelectionFlag.Rows)
                else:
                    super().select(index_or_selection, flag)
            # deselecting has empty list of indexes
            except IndexError:
                super().select(index_or_selection, QItemSelectionModel.SelectionFlag.Clear)
        else:
            if index_or_selection.isValid():
                super().select(index_or_selection, flag)
            else:
                self.clear()



import os

from PySide6.QtCore import QPoint, QSize, Qt, QUrl
from PySide6.QtGui import QIntValidator, QMouseEvent, QDesktopServices
from PySide6.QtWidgets import QAbstractItemView, QDialog
from PySide6.QtWidgets import QGridLayout, QHBoxLayout, QLineEdit
from PySide6.QtWidgets import QMessageBox, QSpacerItem, QSplitter, QTableView
from PySide6.QtWidgets import QVBoxLayout

from OSCR import LiveParser

from .callbacks import (
        auto_split_callback, combat_split_callback, copy_live_data_callback, trim_logfile,
        repair_logfile)
from .displayer import create_live_graph, update_live_display, update_live_graph, update_live_table
from .datamodels import LiveParserTableModel
from .headers import get_live_table_headers
from .style import get_style, get_style_class, theme_font
from .textedit import format_path
from .widgetbuilder import create_button, create_frame, create_icon_button, create_label
from .widgetbuilder import ABOTTOM, AHCENTER, ALEFT, ARIGHT, AVCENTER, RFIXED
from .widgetbuilder import SEXPAND, SMAX, SMAXMAX, SMINMAX, SMINMIN
from .widgets import FlipButton, LiveParserWindow, SizeGrip

import gettext

_ = gettext.gettext


def show_warning(self, title: str, message: str):
    """
    Displays a warning in form of a message box

    Parameters:
    - :param title: title of the warning
    - :param message: message to be displayed
    """
    error = QMessageBox()
    error.setIcon(QMessageBox.Icon.Warning),
    error.setText(message)
    error.setWindowTitle(title)
    error.setStandardButtons(QMessageBox.StandardButton.Ok)
    error.setWindowIcon(self.icons['oscr'])
    error.exec()


def log_size_warning(self, translate):
    """
    Warns user about oversized logfile.
    Note: The default button counts as a two buttons

    :return: "cancel", "split dialog", "continue"
    """

    if self._ is None:
        self._ = translate

    dialog = QMessageBox()
    dialog.setIcon(QMessageBox.Icon.Warning)
    message = (
            self._('The combatlog file you are trying to open will impair the performance of the app ') +
            self._('due to its size. It is advised to split the log. \n\nClick "Split Dialog" to split ') +
            self._('the file, "Cancel" to abort combatlog analysis or "Continue" to analyze the log ') +
            self._('nevertheless.'))
    dialog.setText(message)
    dialog.setWindowTitle('Open Source Combalog Reader')
    dialog.setWindowIcon(self.icons['oscr'])

    dialog.addButton(self._('Continue'), QMessageBox.ButtonRole.AcceptRole)
    default_button = dialog.addButton(self._('Split Dialog'), QMessageBox.ButtonRole.ActionRole)
    dialog.addButton(self._('Trim'), QMessageBox.ButtonRole.ActionRole)
    dialog.addButton(self._('Cancel'), QMessageBox.ButtonRole.RejectRole)

    dialog.setDefaultButton(default_button)
    clicked = dialog.exec()

    if clicked == 1:
        return 'split dialog'
    elif clicked == 2:
        return 'continue'
    elif clicked == 3:
        return 'split dialog'
    elif clicked == 4:
        return 'trim'

    return 'cancel'


def split_dialog(self, translate):
    """
    Opens dialog to split the current logfile.
    """

    if self._ is None:
        self._ = translate

    main_layout = QVBoxLayout()
    thick = self.theme['app']['frame_thickness']
    item_spacing = self.theme['defaults']['isp']
    main_layout.setContentsMargins(thick, thick, thick, thick)
    content_frame = create_frame(self)
    main_layout.addWidget(content_frame)
    current_logpath = self.entry.text()
    vertical_layout = QVBoxLayout()
    vertical_layout.setContentsMargins(thick, thick, thick, thick)
    vertical_layout.setSpacing(item_spacing)
    log_layout = QHBoxLayout()
    log_layout.setContentsMargins(0, 0, 0, 0)
    log_layout.setSpacing(item_spacing)
    current_log_heading = create_label(self, self._('Selected Logfile:'), 'label_subhead')
    log_layout.addWidget(current_log_heading, alignment=ALEFT)
    current_log_label = create_label(self, format_path(current_logpath), 'label')
    log_layout.addWidget(current_log_label, alignment=AVCENTER)
    log_layout.addSpacerItem(QSpacerItem(1, 1, hData=SEXPAND, vData=SMAX))
    vertical_layout.addLayout(log_layout)
    seperator_1 = create_frame(self, content_frame, 'hr', size_policy=SMINMIN)
    seperator_1.setFixedHeight(self.theme['hr']['height'])
    vertical_layout.addWidget(seperator_1)
    grid_layout = QGridLayout()
    grid_layout.setContentsMargins(0, 0, 0, 0)
    grid_layout.setVerticalSpacing(0)
    grid_layout.setHorizontalSpacing(item_spacing)
    vertical_layout.addLayout(grid_layout)

    trim_heading = create_label(self, self._('Trim Logfile:'), 'label_heading')
    grid_layout.addWidget(trim_heading, 0, 0, alignment=ALEFT)
    label_text = (
            self._('Removes all combats but the most recent one from the selected logfile. ') +
            self._('All previous combats will be lost!'))
    trim_text = create_label(self, label_text, 'label')
    trim_text.setWordWrap(True)
    trim_text.setFixedWidth(self.sidebar_item_width)
    grid_layout.addWidget(trim_text, 1, 0, alignment=ALEFT)
    trim_button = create_button(self, self._('Trim'))
    trim_button.clicked.connect(lambda: trim_logfile(self))
    grid_layout.addWidget(trim_button, 1, 2, alignment=ARIGHT | ABOTTOM)
    grid_layout.setRowMinimumHeight(2, item_spacing)
    seperator_3 = create_frame(self, content_frame, 'hr', size_policy=SMINMIN)
    seperator_3.setFixedHeight(self.theme['hr']['height'])
    grid_layout.addWidget(seperator_3, 3, 0, 1, 3)
    grid_layout.setRowMinimumHeight(4, item_spacing)

    auto_split_heading = create_label(self, self._('Split Log Automatically:'), 'label_heading')
    grid_layout.addWidget(auto_split_heading, 5, 0, alignment=ALEFT)
    label_text = (
            self._('Automatically splits the logfile at the next combat end after ') +
            f'{self.settings.value("split_log_after", type=int):,}'+
            self._(' lines until the entire file has ') +
            self._(' been split. The new files are written to the selected folder. It is advised to ') +
            self._('select an empty folder to ensure all files are saved correctly.'))
    auto_split_text = create_label(self, label_text, 'label')
    auto_split_text.setWordWrap(True)
    auto_split_text.setFixedWidth(self.sidebar_item_width)
    grid_layout.addWidget(auto_split_text, 6, 0, alignment=ALEFT)
    auto_split_button = create_button(self, self._('Auto Split'))
    auto_split_button.clicked.connect(lambda: auto_split_callback(self, current_logpath))
    grid_layout.addWidget(auto_split_button, 6, 2, alignment=ARIGHT | ABOTTOM)
    grid_layout.setRowMinimumHeight(7, item_spacing)
    seperator_8 = create_frame(self, content_frame, 'hr', size_policy=SMINMIN)
    seperator_8.setFixedHeight(self.theme['hr']['height'])
    grid_layout.addWidget(seperator_8, 8, 0, 1, 3)
    grid_layout.setRowMinimumHeight(9, item_spacing)
    range_split_heading = create_label(self, self._('Export Range of Combats:'), 'label_heading')
    grid_layout.addWidget(range_split_heading, 10, 0, alignment=ALEFT)
    label_text = (
            self._('Exports combats including and between lower and upper limit to selected file. ') +
            self._('Both limits refer to the indexed list of all combats in the file starting with 1. ') +
            self._('An upper limit larger than the total number of combats or of "-1", is treated as ') +
            self._('being equal to the total number of combats.'))
    range_split_text = create_label(self, label_text, 'label')
    range_split_text.setWordWrap(True)
    range_split_text.setFixedWidth(self.sidebar_item_width)
    grid_layout.addWidget(range_split_text, 11, 0, alignment=ALEFT)
    range_limit_layout = QGridLayout()
    range_limit_layout.setContentsMargins(0, 0, 0, 0)
    range_limit_layout.setSpacing(0)
    range_limit_layout.setRowStretch(0, 1)
    lower_range_label = create_label(self, self._('Lower Limit:'), 'label')
    range_limit_layout.addWidget(lower_range_label, 1, 0, alignment=AVCENTER)
    upper_range_label = create_label(self, self._('Upper Limit:'), 'label')
    range_limit_layout.addWidget(upper_range_label, 2, 0, alignment=AVCENTER)
    lower_range_entry = QLineEdit()
    lower_validator = QIntValidator()
    lower_validator.setBottom(1)
    lower_range_entry.setValidator(lower_validator)
    lower_range_entry.setText('1')
    lower_range_entry.setStyleSheet(
            get_style(self, 'entry', {'margin-top': 0, 'margin-left': '@csp'}))
    lower_range_entry.setFixedWidth(self.sidebar_item_width // 7)
    range_limit_layout.addWidget(lower_range_entry, 1, 1, alignment=AVCENTER)
    upper_range_entry = QLineEdit()
    upper_validator = QIntValidator()
    upper_validator.setBottom(-1)
    upper_range_entry.setValidator(upper_validator)
    upper_range_entry.setText('1')
    upper_range_entry.setStyleSheet(
            get_style(self, 'entry', {'margin-top': 0, 'margin-left': '@csp'}))
    upper_range_entry.setFixedWidth(self.sidebar_item_width // 7)
    range_limit_layout.addWidget(upper_range_entry, 2, 1, alignment=AVCENTER)
    grid_layout.addLayout(range_limit_layout, 11, 1)
    range_split_button = create_button(self, self._('Export Combats'))
    range_split_button.clicked.connect(
            lambda le=lower_range_entry, ue=upper_range_entry:
            combat_split_callback(self, current_logpath, le.text(), ue.text()))
    grid_layout.addWidget(range_split_button, 11, 2, alignment=ARIGHT | ABOTTOM)
    grid_layout.setRowMinimumHeight(12, item_spacing)
    seperator_13 = create_frame(self, content_frame, 'hr', size_policy=SMINMIN)
    seperator_13.setFixedHeight(self.theme['hr']['height'])
    grid_layout.addWidget(seperator_13, 13, 0, 1, 3)
    grid_layout.setRowMinimumHeight(14, item_spacing)
    repair_log_heading = create_label(self, 'Repair Logfile', 'label_heading')
    grid_layout.addWidget(repair_log_heading, 15, 0, alignment=ALEFT)
    repair_log_button = create_button(self, 'Repair')
    repair_log_button.clicked.connect(lambda: repair_logfile(self))
    grid_layout.addWidget(repair_log_button, 16, 2, alignment=ARIGHT | ABOTTOM)

    content_frame.setLayout(vertical_layout)

    dialog = QDialog(self.window)
    dialog.setLayout(main_layout)
    dialog.setWindowTitle(self._('OSCR - Split Logfile'))
    dialog.setStyleSheet(get_style(self, 'dialog_window'))
    dialog.setSizePolicy(SMAXMAX)
    dialog.exec()

def uploadresult_view_result(result):
    """
    Open the result up in the user's web browser.

    Parameters:
    - :param result: The Combat Log Upload Response
    """
    from .leagueconnector import OSCR_SERVER_BACKEND

    url = QUrl(f"{OSCR_SERVER_BACKEND}/ui/combatlog/{result.combatlog}/")
    QDesktopServices.openUrl(url)


def uploadresult_dialog(self, result, translate):
    """
    Shows a dialog that informs about the result of the triggered upload.

    Paramters:
    - :param result: dict containing result
    """

    if self._ is None:
        self._ = translate

    dialog = QDialog(self.window)
    main_layout = QVBoxLayout()
    thick = self.theme['app']['frame_thickness']
    main_layout.setContentsMargins(thick, thick, thick, thick)
    content_frame = create_frame(self)
    main_layout.addWidget(content_frame)
    content_layout = QGridLayout()
    content_layout.setContentsMargins(thick, thick, thick, thick)
    content_layout.setSpacing(0)
    margin = {'margin-bottom': self.theme['defaults']['isp']}
    title_label = create_label(self, f"{result.detail}", 'label_heading', style_override=margin)
    content_layout.addWidget(title_label, 0, 0, 1, 4, alignment=ALEFT)
    view_button = create_button(self, 'View Online', style_override=margin)
    view_button.clicked.connect(lambda: uploadresult_view_result(result))
    if result.results:
        content_layout.addWidget(view_button, 0, 0, 1, 4, alignment=ARIGHT)
    icon_size = QSize(self.config['icon_size'] / 1.5, self.config['icon_size'] / 1.5)
    row = 0
    if (result.results):
        for row, line in enumerate(result.results, 1):
            if row % 2 == 1:
                table_style = {'background-color': '@mbg', 'padding': (5, 3, 3, 3), 'margin': 0}
                icon_table_style = {'background-color': '@mbg', 'padding': (3, 3, 3, 3), 'margin': 0}
            else:
                table_style = {'background-color': '@bg', 'padding': (5, 3, 3, 3), 'margin': 0}
                icon_table_style = {'background-color': '@bg', 'padding': (3, 3, 3, 3), 'margin': 0}
            if line.updated:
                icon = self.icons['check'].pixmap(icon_size)
            else:
                icon = self.icons['dash'].pixmap(icon_size)
            status_label = create_label(self, '', style_override=icon_table_style)
            status_label.setPixmap(icon)
            status_label.setSizePolicy(SMINMIN)
            content_layout.addWidget(status_label, row, 0)
            name_label = create_label(self, line.name, style_override=table_style)
            name_label.setSizePolicy(SMINMAX)
            content_layout.addWidget(name_label, row, 1)
            value_label = create_label(self, str(line.value), style_override=table_style)
            value_label.setSizePolicy(SMINMAX)
            value_label.setAlignment(ARIGHT)
            content_layout.addWidget(value_label, row, 2)
            detail_label = create_label(self, line.detail, style_override=table_style)
            detail_label.setSizePolicy(SMINMAX)
            content_layout.addWidget(detail_label, row, 3)
    top_margin = {'margin-top': self.theme['defaults']['isp']}
    close_button = create_button(self, 'Close', style_override=top_margin)
    close_button.clicked.connect(dialog.close)
    content_layout.addWidget(close_button, row + 1, 0, 1, 4, alignment=AHCENTER)
    content_frame.setLayout(content_layout)

    dialog.setLayout(main_layout)
    dialog.setWindowTitle(self._('OSCR - Upload Results'))
    dialog.setStyleSheet(get_style(self, 'dialog_window'))
    dialog.setSizePolicy(SMAXMAX)
    dialog.setFixedSize(dialog.sizeHint())
    dialog.exec()


def live_parser_toggle(self, translate, activate):
    """
    Activates / Deactivates LiveParser.

    Parameters:
    - :param activate: True when parser should be shown; False when open parser should be closed.
    """

    if self._ is None:
        self._ = translate

    if activate:
        log_path = self.settings.value('sto_log_path')
        if not log_path or not os.path.isfile(log_path):
            show_warning(
                    self, self._('Invalid Logfile'), self._('Make sure to set the STO Logfile setting in the ') +
                    self._('settings tab to a valid logfile before starting the live parser.'))
            self.widgets.live_parser_button.setChecked(False)
            return
        FIELD_INDEX_CONVERSION = {0: 0, 1: 2, 2: 3, 3: 4}
        graph_active = self.settings.value('live_graph_active', type=bool)
        data_buffer = []
        data_field = FIELD_INDEX_CONVERSION[self.settings.value('live_graph_field', type=int)]
        self.live_parser = LiveParser(log_path, update_callback=lambda data: update_live_display(
                self, data, graph_active, data_buffer, data_field),
                settings=self.live_parser_settings)
        create_live_parser_window(self, translate)
    else:
        try:
            self.live_parser_window.close()
        except AttributeError:
            pass
        try:
            self.live_parser.stop()
        except AttributeError:
            pass
        self.live_parser_window.update_table.disconnect()
        self.live_parser_window.update_graph.disconnect()
        self.live_parser_window.deleteLater()
        self.live_parser_window = None
        self.live_parser = None
        self.widgets.live_parser_table = None
        self.widgets.live_parser_splitter = None
        self.widgets.live_parser_button.setChecked(False)


def create_live_parser_window(self, translate):
    """
    Creates the LiveParser window.
    """
    ui_scale = self.config['ui_scale']
    self.config['ui_scale'] = self.config['live_scale']

    live_window = LiveParserWindow()
    live_window.setStyleSheet(get_style(self, 'live_parser'))
    live_window.setWindowFlags(
            live_window.windowFlags()
            | Qt.WindowType.WindowStaysOnTopHint
            | Qt.WindowType.WindowDoesNotAcceptFocus
            | Qt.WindowType.SubWindow
            | Qt.WindowType.FramelessWindowHint)
    # live_window.setAttribute(Qt.WidgetAttribute.WA_ShowWithoutActivating, True)
    live_window.setWindowOpacity(self.settings.value('live_parser_opacity', type=float))
    if self.settings.value('live_geometry'):
        live_window.restoreGeometry(self.settings.value('live_geometry'))
    live_window.closeEvent = lambda close_event: live_parser_close_callback(self, close_event)
    live_window.mousePressEvent = lambda press_event: live_parser_press_event(self, press_event)
    live_window.mouseMoveEvent = lambda move_event: live_parser_move_event(self, move_event)
    live_window.setSizePolicy(SMAXMAX)
    layout = QVBoxLayout()
    layout.setContentsMargins(0, 0, 0, 0)
    layout.setSpacing(0)

    graph_colors = None
    graph_column = None
    graph_active = self.settings.value('live_graph_active', type=bool)
    if graph_active:
        splitter = QSplitter(Qt.Orientation.Vertical)
        splitter.setStyleSheet(get_style_class(self, 'QSplitter', 'splitter'))
        splitter.setChildrenCollapsible(False)
        self.widgets.live_parser_splitter = splitter
        graph_frame, curves = create_live_graph(self)
        graph_frame.setMinimumHeight(self.sidebar_item_width * 0.1)
        splitter.addWidget(graph_frame)
        self.widgets.live_parser_curves = curves
        FIELD_INDEX_CONVERSION = {0: 0, 1: 2, 2: 3, 3: 4}
        graph_column = FIELD_INDEX_CONVERSION[self.settings.value('live_graph_field', type=int)]
        graph_colors = self.theme['plot']['color_cycler'][:5]
        layout.addWidget(splitter, stretch=1)

    table = QTableView()
    table.setAlternatingRowColors(self.theme['s.c']['table_alternate'])
    table.setShowGrid(self.theme['s.c']['table_gridline'])
    table.setStyleSheet(get_style_class(self, 'QTableView', 'live_table'))
    table.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
    table.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
    table.horizontalHeader().setStyleSheet(
            get_style_class(self, 'QHeaderView', 'live_table_header'))
    table.verticalHeader().setStyleSheet(get_style_class(self, 'QHeaderView', 'live_table_index'))
    table.verticalHeader().setMinimumHeight(1)
    table.verticalHeader().setDefaultSectionSize(1)
    table.horizontalHeader().setMinimumWidth(1)
    table.horizontalHeader().setDefaultSectionSize(1)
    table.horizontalHeader().setSectionResizeMode(RFIXED)
    table.verticalHeader().setSectionResizeMode(RFIXED)
    table.setSizePolicy(SMINMIN)
    table.setSelectionMode(QAbstractItemView.SelectionMode.NoSelection)
    table.setMinimumWidth(self.sidebar_item_width * 0.1)
    table.setMinimumHeight(self.sidebar_item_width * 0.1)
    model = LiveParserTableModel(
            [[0] * len(get_live_table_headers())], get_live_table_headers(), [''],
            theme_font(self, 'live_table_header'), theme_font(self, 'live_table'),
            legend_col=graph_column, colors=graph_colors)
    table.setModel(model)
    table.resizeColumnsToContents()
    table.resizeRowsToContents()
    for index in range(len(get_live_table_headers())):
        if not self.settings.value(f'live_columns|{index}', type=bool):
            table.hideColumn(index)
    self.widgets.live_parser_table = table
    if graph_active:
        splitter.addWidget(table)
        if self.settings.value('live_splitter'):
            splitter.restoreState(self.settings.value('live_splitter'))
    else:
        layout.addWidget(table, 1)

    bottom_layout = QGridLayout()
    bottom_layout.setContentsMargins(self.theme['defaults']['isp'], 0, 0, 0)
    bottom_layout.setSpacing(0)
    bottom_layout.setColumnStretch(0, 1)
    bottom_layout.setColumnStretch(2, 1)

    icon_size = [self.theme['s.c']['button_icon_size'] * self.config['live_scale'] * 0.8] * 2
    copy_button = copy_button = create_icon_button(
            self, self.icons['copy'], self._('Copy Result'), icon_size=icon_size)
    copy_button.clicked.connect(lambda: copy_live_data_callback(self))
    bottom_layout.addWidget(copy_button, 0, 0, alignment=ARIGHT | AVCENTER)
    activate_button = FlipButton(self._('Activate'), self._('Deactivate'), live_window, checkable=True)
    activate_button.setStyleSheet(self.get_style_class(
            'QPushButton', 'toggle_button', {'margin': (0, 8, 0, 8)}))
    activate_button.setFont(self.theme_font('app', '@subhead'))
    activate_button.r_function = lambda: self.live_parser.start()
    activate_button.l_function = lambda: self.live_parser.stop()
    bottom_layout.addWidget(activate_button, 0, 1, alignment=AVCENTER)
    close_button = create_icon_button(
            self, self.icons['close'], self._('Close Live Parser'), icon_size=icon_size)
    close_button.clicked.connect(lambda: live_parser_toggle(self, translate, False))
    bottom_layout.addWidget(close_button, 0, 2, alignment=ALEFT | AVCENTER)

    grip = SizeGrip(live_window)
    grip.setStyleSheet(get_style(self, 'resize_handle'))
    bottom_layout.addWidget(grip, 0, 3, alignment=ARIGHT | ABOTTOM)

    layout.addLayout(bottom_layout)
    live_window.setLayout(layout)
    live_window.update_table.connect(lambda data: update_live_table(self, data))
    live_window.update_graph.connect(update_live_graph)
    self.live_parser_window = live_window
    self.config['ui_scale'] = ui_scale
    live_window.show()

    if self.settings.value('live_enabled', type=bool):
        activate_button.flip()


def live_parser_close_callback(self, event):
    """
    Executed when application is closed.
    """
    window_geometry = self.live_parser_window.saveGeometry()
    self.settings.setValue('live_geometry', window_geometry)
    try:
        self.settings.setValue('live_splitter', self.widgets.live_parser_splitter.saveState())
    except AttributeError:
        pass
    event.accept()


def live_parser_press_event(self, event: QMouseEvent):
    self.live_parser_window.start_pos = event.globalPosition().toPoint()
    event.accept()


def live_parser_move_event(self, event: QMouseEvent):
    parser_window = self.live_parser_window
    pos_delta = QPoint(event.globalPosition().toPoint() - parser_window.start_pos)
    parser_window.move(parser_window.x() + pos_delta.x(), parser_window.y() + pos_delta.y())
    parser_window.start_pos = event.globalPosition().toPoint()
    event.accept()



import gettext
import os

def init_translation(lang_code='en'):
    """
    Initialize translation.
    :param lang_code: Language codes, Example: 'en', 'zh', 'fr'
    :return: gettext translation function
    """
    try:
        lang = gettext.translation('messages', localedir='locales', languages=[lang_code], fallback=True)
        lang.install()
    except Exception as e:
        print(e)

    return lang.gettext
